# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import logging
from datetime import datetime
from typing import Any

from oagi import AsyncActor
from oagi.constants import (
    DEFAULT_MAX_STEPS,
    DEFAULT_REFLECTION_INTERVAL,
    DEFAULT_STEP_DELAY,
    DEFAULT_TEMPERATURE,
    MODEL_ACTOR,
)
from oagi.handler.utils import configure_handler_delay, reset_handler
from oagi.types import (
    URL,
    ActionEvent,
    AsyncActionHandler,
    AsyncImageProvider,
    AsyncObserver,
    Image,
    PlanEvent,
    StepEvent,
    extract_uuid_from_url,
)

from ..protocol import AsyncAgent
from .memory import PlannerMemory
from .models import Action, ExecutionResult
from .planner import Planner

logger = logging.getLogger(__name__)


def _serialize_image(image: Image | str) -> bytes | str:
    """Convert an image to bytes or keep URL as string."""
    if isinstance(image, str):
        return image
    return image.read()


class TaskeeAgent(AsyncAgent):
    """Executes a single todo with planning and reflection capabilities.

    This agent uses a Planner to:
    1. Convert a todo into a clear actionable instruction
    2. Execute the instruction using OAGI API
    3. Periodically reflect on progress and adjust approach
    4. Generate execution summaries
    """

    def __init__(
        self,
        api_key: str | None = None,
        base_url: str | None = None,
        model: str = MODEL_ACTOR,
        max_steps: int = DEFAULT_MAX_STEPS,
        reflection_interval: int = DEFAULT_REFLECTION_INTERVAL,
        temperature: float = DEFAULT_TEMPERATURE,
        planner: Planner | None = None,
        external_memory: PlannerMemory | None = None,
        todo_index: int | None = None,
        step_observer: AsyncObserver | None = None,
        step_delay: float = DEFAULT_STEP_DELAY,
    ):
        """Initialize the taskee agent.

        Args:
            api_key: OAGI API key
            base_url: OAGI API base URL
            model: Model to use for vision tasks
            max_steps: Maximum steps before reinitializing task
            reflection_interval: Number of actions before triggering reflection
            temperature: Sampling temperature
            planner: Planner for planning and reflection
            external_memory: External memory from parent agent
            todo_index: Index of the todo being executed
            step_observer: Optional observer for step tracking
            step_delay: Delay in seconds after actions before next screenshot
        """
        self.api_key = api_key
        self.base_url = base_url
        self.model = model
        self.max_steps = max_steps
        self.reflection_interval = reflection_interval
        self.temperature = temperature
        self.planner = planner or Planner(api_key=api_key, base_url=base_url)
        self.external_memory = external_memory
        self.todo_index = todo_index
        self.step_observer = step_observer
        self.step_delay = step_delay

        # Internal state
        self.actor: AsyncActor | None = None
        self.current_todo: str = ""
        self.current_instruction: str = ""
        self.actions: list[Action] = []
        self.total_actions = 0
        self.since_reflection = 0
        self.success = False

    async def execute(
        self,
        instruction: str,
        action_handler: AsyncActionHandler,
        image_provider: AsyncImageProvider,
    ) -> bool:
        """Execute the todo using planning and reflection.

        Args:
            instruction: The todo description to execute
            action_handler: Handler for executing actions
            image_provider: Provider for capturing screenshots

        Returns:
            True if successful, False otherwise
        """
        # Reset handler state at todo execution start
        reset_handler(action_handler)

        # Configure handler's post_batch_delay from agent's step_delay
        configure_handler_delay(action_handler, self.step_delay)

        self.current_todo = instruction
        self.actions = []
        self.total_actions = 0
        self.since_reflection = 0
        self.success = False

        try:
            self.actor = AsyncActor(
                api_key=self.api_key,
                base_url=self.base_url,
                model=self.model,
                temperature=self.temperature,
            )
            # Initial planning
            await self._initial_plan(image_provider)

            # Initialize the actor with the task
            await self.actor.init_task(
                self.current_instruction, max_steps=self.max_steps
            )

            # Main execution loop with reinitializations
            remaining_steps = self.max_steps

            while remaining_steps > 0 and not self.success:
                # Execute subtask
                steps_taken = await self._execute_subtask(
                    min(self.max_steps, remaining_steps),
                    action_handler,
                    image_provider,
                )
                remaining_steps -= steps_taken

                # Check if we should continue
                if not self.success and remaining_steps > 0:
                    # Reflect and potentially get new instruction
                    should_continue = await self._reflect_and_decide(image_provider)
                    if not should_continue:
                        break

            # Generate final summary
            await self._generate_summary()

            return self.success

        except Exception as e:
            logger.error(f"Error executing todo: {e}")
            self._record_action(
                action_type="error",
                target=None,
                reasoning=str(e),
            )
            return False
        finally:
            # Clean up actor
            if self.actor:
                await self.actor.close()
                self.actor = None

    async def _initial_plan(self, image_provider: AsyncImageProvider) -> None:
        """Generate initial plan for the todo.

        Args:
            image_provider: Provider for capturing screenshots
        """
        logger.info("Generating initial plan for todo")

        # Capture initial screenshot
        screenshot = await image_provider()

        # Get context from external memory if available
        context = self._get_context()

        # Generate plan using LLM planner
        plan_output, request_id = await self.planner.initial_plan(
            self.current_todo,
            context,
            screenshot,
            memory=self.external_memory,
            todo_index=self.todo_index,
        )

        # Record planning action
        self._record_action(
            action_type="plan",
            target="initial",
            reasoning=plan_output.reasoning,
            result=plan_output.instruction,
        )

        # Emit plan event
        if self.step_observer:
            await self.step_observer.on_event(
                PlanEvent(
                    phase="initial",
                    image=_serialize_image(screenshot),
                    reasoning=plan_output.reasoning,
                    result=plan_output.instruction,
                    request_id=request_id,
                )
            )

        # Set current instruction
        self.current_instruction = plan_output.instruction
        logger.info(f"Initial instruction: {self.current_instruction}")

        # Handle subtodos if any
        if plan_output.subtodos:
            logger.info(f"Planner created {len(plan_output.subtodos)} subtodos")
            # Could potentially add these to memory for tracking

    async def _execute_subtask(
        self,
        max_steps: int,
        action_handler: AsyncActionHandler,
        image_provider: AsyncImageProvider,
    ) -> int:
        """Execute a subtask with the current instruction.

        Args:
            max_steps: Maximum steps for this subtask
            action_handler: Handler for executing actions
            image_provider: Provider for capturing screenshots

        Returns:
            Number of steps taken
        """
        logger.info(f"Executing subtask with max {max_steps} steps")

        steps_taken = 0
        client = self.planner._ensure_client()

        for step_num in range(max_steps):
            # Capture screenshot
            screenshot = await image_provider()

            # Get screenshot UUID - either extract from URL or upload
            try:
                screenshot_uuid = None
                screenshot_url = None

                # Check if screenshot is already a URL (from SocketIOImageProvider)
                if isinstance(screenshot, str):
                    screenshot_uuid = extract_uuid_from_url(screenshot)
                    screenshot_url = screenshot

                # If not a URL or UUID extraction failed, upload the image
                if not screenshot_uuid:
                    upload_response = await client.put_s3_presigned_url(screenshot)
                    screenshot_uuid = upload_response.uuid
                    screenshot_url = upload_response.download_url
            except Exception as e:
                logger.error(f"Error uploading screenshot: {e}")
                self._record_action(
                    action_type="error",
                    target="screenshot_upload",
                    reasoning=str(e),
                )
                break

            # Get next step from OAGI using URL (avoids re-upload)
            try:
                step = await self.actor.step(URL(screenshot_url), instruction=None)
            except Exception as e:
                logger.error(f"Error getting step from OAGI: {e}")
                self._record_action(
                    action_type="error",
                    target="oagi_step",
                    reasoning=str(e),
                    screenshot_uuid=screenshot_uuid,
                )
                break

            # Log reasoning
            if step.reason:
                logger.info(f"Step {self.total_actions + 1}: {step.reason}")

            # Emit step event
            if self.step_observer:
                await self.step_observer.on_event(
                    StepEvent(
                        step_num=self.total_actions + 1,
                        image=_serialize_image(screenshot),
                        step=step,
                        task_id=self.actor.task_id,
                    )
                )

            # Record OAGI actions
            if step.actions:
                # Log actions with details
                logger.info(f"Actions ({len(step.actions)}):")
                for action in step.actions:
                    count_suffix = (
                        f" x{action.count}" if action.count and action.count > 1 else ""
                    )
                    logger.info(
                        f"  [{action.type.value}] {action.argument}{count_suffix}"
                    )

                for action in step.actions:
                    self._record_action(
                        action_type=action.type.lower(),
                        target=action.argument,
                        reasoning=step.reason,
                        screenshot_uuid=screenshot_uuid,
                    )

                # Execute actions
                error = None
                try:
                    await action_handler(step.actions)
                except Exception as e:
                    error = str(e)
                    raise

                # Emit action event
                if self.step_observer:
                    await self.step_observer.on_event(
                        ActionEvent(
                            step_num=self.total_actions + 1,
                            actions=step.actions,
                            error=error,
                        )
                    )

                self.total_actions += len(step.actions)
                self.since_reflection += len(step.actions)

            steps_taken += 1

            # Check if task is complete
            if step.stop:
                logger.info("OAGI signaled task completion")
                break

            # Check if reflection is needed
            if self.since_reflection >= self.reflection_interval:
                logger.info("Reflection interval reached")
                break

        return steps_taken

    async def _reflect_and_decide(self, image_provider: AsyncImageProvider) -> bool:
        """Reflect on progress and decide whether to continue.

        Args:
            image_provider: Provider for capturing screenshots

        Returns:
            True to continue, False to stop
        """
        logger.info("Reflecting on progress")

        # Capture current screenshot
        screenshot = await image_provider()

        # Get context
        context = self._get_context()
        context["current_todo"] = self.current_todo

        # Get recent actions for reflection
        recent_actions = self.actions[-self.since_reflection :]

        # Reflect using planner
        reflection, request_id = await self.planner.reflect(
            recent_actions,
            context,
            screenshot,
            memory=self.external_memory,
            todo_index=self.todo_index,
            current_instruction=self.current_instruction,
            reflection_interval=self.reflection_interval,
        )

        # Record reflection
        self._record_action(
            action_type="reflect",
            target=None,
            reasoning=reflection.reasoning,
            result=("continue" if reflection.continue_current else "pivot"),
        )

        # Emit plan event for reflection
        if self.step_observer:
            decision = (
                "success"
                if reflection.success_assessment
                else ("continue" if reflection.continue_current else "pivot")
            )
            await self.step_observer.on_event(
                PlanEvent(
                    phase="reflection",
                    image=_serialize_image(screenshot),
                    reasoning=reflection.reasoning,
                    result=decision,
                    request_id=request_id,
                )
            )

        # Update success assessment
        if reflection.success_assessment:
            self.success = True
            logger.info("Reflection indicates task is successful")
            return False

        # Reset reflection counter
        self.since_reflection = 0

        # Update instruction if needed
        if not reflection.continue_current and reflection.new_instruction:
            logger.info(f"Pivoting to new instruction: {reflection.new_instruction}")
            self.current_instruction = reflection.new_instruction

            # the following line create a new actor
            await self.actor.init_task(
                self.current_instruction, max_steps=self.max_steps
            )
            return True

        return reflection.continue_current

    async def _generate_summary(self) -> None:
        """Generate execution summary."""
        logger.info("Generating execution summary")

        context = self._get_context()
        context["current_todo"] = self.current_todo

        summary, request_id = await self.planner.summarize(
            self.actions,
            context,
            memory=self.external_memory,
            todo_index=self.todo_index,
        )

        # Record summary
        self._record_action(
            action_type="summary",
            target=None,
            reasoning=summary,
        )

        # Emit plan event for summary
        if self.step_observer:
            await self.step_observer.on_event(
                PlanEvent(
                    phase="summary",
                    image=None,
                    reasoning=summary,
                    result=None,
                    request_id=request_id,
                )
            )

        logger.info(f"Execution summary: {summary}")

    def _record_action(
        self,
        action_type: str,
        target: str | None,
        reasoning: str | None = None,
        result: str | None = None,
        screenshot_uuid: str | None = None,
    ) -> None:
        """Record an action to the history.

        Args:
            action_type: Type of action
            target: Target of the action
            reasoning: Reasoning for the action
            result: Result of the action
            screenshot_uuid: UUID of uploaded screenshot for this action
        """
        action = Action(
            timestamp=datetime.now().isoformat(),
            action_type=action_type,
            target=target,
            reasoning=reasoning,
            result=result,
            details={},
            screenshot_uuid=screenshot_uuid,
        )
        self.actions.append(action)

    def _get_context(self) -> dict[str, Any]:
        """Get execution context.

        Returns:
            Dictionary with context information
        """
        if self.external_memory:
            return self.external_memory.get_context()
        return {}

    def return_execution_results(self) -> ExecutionResult:
        """Return the execution results.

        Returns:
            ExecutionResult with success status, actions, and summary
        """
        # Find summary in actions
        summary = ""
        for action in reversed(self.actions):
            if action.action_type == "summary":
                summary = action.reasoning or ""
                break

        return ExecutionResult(
            success=self.success,
            actions=self.actions,
            summary=summary,
            total_steps=self.total_actions,
        )



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

instruction_template = """You are a Desktop Agent completing computer use tasks from a user instruction.

Every step, you will look at the screenshot and output the desired actions in a format as:

<|think_start|> brief description of your intent and reasoning <|think_end|>
<|action_start|> one of the allowed actions as below <|action_end|>

In the action field, you have the following action formats:
1. click(x, y) # left-click at the position (x, y), where x and y are integers normalized between 0 and 1000
2. left_double(x, y) # left-double-click at the position (x, y), where x and y are integers normalized between 0 and 1000
3. left_triple(x, y) # left-triple-click at the position (x, y), where x and y are integers normalized between 0 and 1000
4. right_single(x, y) # right-click at the position (x, y), where x and y are integers normalized between 0 and 1000
5. drag(x1, y1, x2, y2) # drag the mouse from (x1, y1) to (x2, y2) to select or move contents, where x1, y1, x2, y2 are integers normalized between 0 and 1000
6. hotkey(key, c) # press the key for c times
7. type(text) # type a text string on the keyboard
8. scroll(x, y, direction, c) # scroll the mouse at position (x, y) in the direction of up or down for c times, where x and y are integers normalized between 0 and 1000
9. wait() # wait for a while
10. finish() # indicate the task is finished

Directly output the text beginning with <|think_start|>, no additional text is needed for this scenario.

The user instruction is:
{instruction}
"""


def build_prompt(task_description: str) -> str:
    """Build the instruction prompt for the OAGI model.

    Args:
        task_description: The task description to include in the prompt

    Returns:
        The formatted prompt string with action format documentation
    """
    return instruction_template.format(instruction=task_description)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------
import asyncio

from oagi import (
    AsyncDefaultAgent,
    AsyncPyautoguiActionHandler,
    AsyncScreenshotMaker,
)


def execute_task_auto(task_desc, max_steps=5):
    """Synchronous wrapper for async task execution."""
    return asyncio.run(async_execute_task_auto(task_desc, max_steps))


async def async_execute_task_auto(task_desc, max_steps=5):
    # set OAGI_API_KEY and OAGI_BASE_URL
    # or AsyncDefaultAgent(api_key="your_api_key", base_url="your_base_url")
    agent = AsyncDefaultAgent(max_steps=max_steps)

    # executor = lambda actions: print(actions) for debugging
    action_handler = AsyncPyautoguiActionHandler()
    image_provider = AsyncScreenshotMaker()

    is_completed = await agent.execute(
        task_desc,
        action_handler=action_handler,
        image_provider=image_provider,
    )

    return is_completed, await image_provider.last_image()



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

from uuid import uuid4

from ..constants import (
    DEFAULT_MAX_STEPS,
    MAX_STEPS_ALLOWED_ACTOR,
    MAX_STEPS_ALLOWED_THINKER,
    MODEL_THINKER,
)
from ..logging import get_logger
from ..types import URL, Image, Step
from ..utils.prompt_builder import build_prompt

logger = get_logger("actor.base")


class BaseActor:
    """Base class with shared task management logic for sync/async actors."""

    def __init__(
        self,
        api_key: str | None,
        base_url: str | None,
        model: str,
        temperature: float | None,
    ):
        self.task_id: str = uuid4().hex  # Client-side generated UUID
        self.task_description: str | None = None
        self.model = model
        self.temperature = temperature
        self.message_history: list = []  # OpenAI-compatible message history
        self.max_steps: int = DEFAULT_MAX_STEPS
        self.current_step: int = 0  # Current step counter
        # Client will be set by subclasses
        self.api_key: str | None = None
        self.base_url: str | None = None

    def _validate_max_steps(self, max_steps: int) -> int:
        """Validate and cap max_steps based on model type.

        Args:
            max_steps: Requested maximum number of steps

        Returns:
            Validated max_steps (capped to model limit if exceeded)
        """
        limit = (
            MAX_STEPS_ALLOWED_THINKER
            if self.model == MODEL_THINKER
            else MAX_STEPS_ALLOWED_ACTOR
        )
        if max_steps > limit:
            logger.warning(
                f"max_steps ({max_steps}) exceeds limit for model '{self.model}'. "
                f"Capping to {limit}."
            )
            return limit
        return max_steps

    def _prepare_init_task(
        self,
        task_desc: str,
        max_steps: int,
    ):
        """Prepare task initialization.

        Args:
            task_desc: Task description
            max_steps: Maximum number of steps
        """
        self.task_id = uuid4().hex
        self.task_description = task_desc
        self.message_history = []
        self.max_steps = self._validate_max_steps(max_steps)
        self.current_step = 0
        logger.info(f"Task initialized: '{task_desc}' (max_steps: {self.max_steps})")

    def _validate_and_increment_step(self):
        if not self.task_description:
            raise ValueError("Task description must be set. Call init_task() first.")
        if self.current_step >= self.max_steps:
            raise ValueError(
                f"Max steps limit ({self.max_steps}) reached. "
                "Call init_task() to start a new task."
            )
        self.current_step += 1

    def _get_temperature(self, temperature: float | None) -> float | None:
        return temperature if temperature is not None else self.temperature

    def _prepare_screenshot(self, screenshot: Image | bytes) -> bytes:
        if isinstance(screenshot, Image):
            return screenshot.read()
        return screenshot

    def _get_screenshot_url(self, screenshot: Image | URL | bytes) -> str | None:
        """Get screenshot URL if it's a string, otherwise return None."""
        if isinstance(screenshot, str):
            return screenshot
        return None

    def _ensure_screenshot_url_sync(
        self, screenshot: Image | URL | bytes, client
    ) -> str:
        """Get screenshot URL, uploading to S3 if needed (sync version).

        Args:
            screenshot: Screenshot as Image object, URL string, or raw bytes
            client: SyncClient instance for S3 upload

        Returns:
            Screenshot URL (either direct or from S3 upload)
        """
        screenshot_url = self._get_screenshot_url(screenshot)
        if screenshot_url is None:
            screenshot_bytes = self._prepare_screenshot(screenshot)
            upload_response = client.put_s3_presigned_url(screenshot_bytes)
            screenshot_url = upload_response.download_url
        return screenshot_url

    async def _ensure_screenshot_url_async(
        self, screenshot: Image | URL | bytes, client
    ) -> str:
        """Get screenshot URL, uploading to S3 if needed (async version).

        Args:
            screenshot: Screenshot as Image object, URL string, or raw bytes
            client: AsyncClient instance for S3 upload

        Returns:
            Screenshot URL (either direct or from S3 upload)
        """
        screenshot_url = self._get_screenshot_url(screenshot)
        if screenshot_url is None:
            screenshot_bytes = self._prepare_screenshot(screenshot)
            upload_response = await client.put_s3_presigned_url(screenshot_bytes)
            screenshot_url = upload_response.download_url
        return screenshot_url

    def _add_user_message_to_history(
        self, screenshot_url: str, prompt: str | None = None
    ):
        """Add user message with screenshot to message history.

        Args:
            screenshot_url: URL of the screenshot
            prompt: Optional prompt text (for first message only)
        """
        content = []
        if prompt:
            content.append({"type": "text", "text": prompt})
        content.append({"type": "image_url", "image_url": {"url": screenshot_url}})

        self.message_history.append(
            {
                "role": "user",
                "content": content,
            }
        )

    def _add_assistant_message_to_history(self, raw_output: str):
        """Add assistant response to message history.

        Args:
            raw_output: Raw model output string
        """
        if raw_output:
            self.message_history.append(
                {
                    "role": "assistant",
                    "content": raw_output,
                }
            )

    def _build_step_prompt(self) -> str | None:
        """Build prompt for first message only."""
        if len(self.message_history) == 0:
            return build_prompt(self.task_description)
        return None

    def _log_step_completion(self, step: Step, prefix: str = "") -> None:
        """Log step completion status."""
        if step.stop:
            logger.info(f"{prefix}Task completed.")
        else:
            logger.debug(f"{prefix}Step completed with {len(step.actions)} actions")

    def _log_step_execution(self, prefix: str = ""):
        logger.debug(f"Executing {prefix}step for task: '{self.task_description}'")

    def _handle_step_error(self, error: Exception, prefix: str = ""):
        logger.error(f"Error during {prefix}step execution: {error}")
        raise


class BaseAutoMode:
    """Base class with shared auto_mode logic for ShortTask implementations."""

    def _log_auto_mode_start(self, task_desc: str, max_steps: int, prefix: str = ""):
        logger.info(
            f"Starting {prefix}auto mode for task: '{task_desc}' (max_steps: {max_steps})"
        )

    def _log_auto_mode_step(self, step_num: int, max_steps: int, prefix: str = ""):
        logger.debug(f"{prefix.capitalize()}auto mode step {step_num}/{max_steps}")

    def _log_auto_mode_actions(self, action_count: int, prefix: str = ""):
        verb = "asynchronously" if "async" in prefix else ""
        logger.debug(f"Executing {action_count} actions {verb}".strip())

    def _log_auto_mode_completion(self, steps: int, prefix: str = ""):
        logger.info(
            f"{prefix.capitalize()}auto mode completed successfully after {steps} steps"
        )

    def _log_auto_mode_max_steps(self, max_steps: int, prefix: str = ""):
        logger.warning(
            f"{prefix.capitalize()}auto mode reached max steps ({max_steps}) without completion"
        )



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import re
from enum import Enum

from pydantic import BaseModel, Field


class ActionType(str, Enum):
    CLICK = "click"
    LEFT_DOUBLE = "left_double"
    LEFT_TRIPLE = "left_triple"
    RIGHT_SINGLE = "right_single"
    DRAG = "drag"
    HOTKEY = "hotkey"
    TYPE = "type"
    SCROLL = "scroll"
    FINISH = "finish"
    WAIT = "wait"
    CALL_USER = "call_user"


class Action(BaseModel):
    type: ActionType = Field(..., description="Type of action to perform")
    argument: str = Field(..., description="Action argument in the specified format")
    count: int | None = Field(
        default=1, ge=1, description="Number of times to repeat the action"
    )


def parse_coords(args_str: str) -> tuple[int, int] | None:
    """Extract x, y coordinates from argument string.

    Args:
        args_str: Argument string in format "x, y" (normalized 0-1000 range)

    Returns:
        Tuple of (x, y) coordinates, or None if parsing fails
    """
    match = re.match(r"(\d+),\s*(\d+)", args_str)
    if not match:
        return None
    return int(match.group(1)), int(match.group(2))


def parse_drag_coords(args_str: str) -> tuple[int, int, int, int] | None:
    """Extract x1, y1, x2, y2 coordinates from drag argument string.

    Args:
        args_str: Argument string in format "x1, y1, x2, y2" (normalized 0-1000 range)

    Returns:
        Tuple of (x1, y1, x2, y2) coordinates, or None if parsing fails
    """
    match = re.match(r"(\d+),\s*(\d+),\s*(\d+),\s*(\d+)", args_str)
    if not match:
        return None
    return (
        int(match.group(1)),
        int(match.group(2)),
        int(match.group(3)),
        int(match.group(4)),
    )


def parse_scroll(args_str: str) -> tuple[int, int, str] | None:
    """Extract x, y, direction from scroll argument string.

    Args:
        args_str: Argument string in format "x, y, direction" (normalized 0-1000 range)

    Returns:
        Tuple of (x, y, direction) where direction is "up" or "down", or None if parsing fails
    """
    match = re.match(r"(\d+),\s*(\d+),\s*(\w+)", args_str)
    if not match:
        return None
    direction = match.group(3).lower()
    if direction not in ("up", "down"):
        return None
    return int(match.group(1)), int(match.group(2)), direction



import re
from typing import NewType

URL = NewType("URL", str)

# Pattern to extract UUID from S3 URL
# Formats:
#   - https://bucket.s3.amazonaws.com/{user_id}/{uuid}.{ext}  (download URL)
#   - https://bucket.s3.amazonaws.com/{user_id}/{uuid}?...    (presigned URL)
_UUID_PATTERN = re.compile(
    r"/([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})(?:\.[a-z]+)?(?:\?|$)",
    re.IGNORECASE,
)


def extract_uuid_from_url(url: str) -> str | None:
    """Extract UUID from S3 URL.

    Args:
        url: S3 URL in one of these formats:
            - https://bucket.s3.amazonaws.com/{user_id}/{uuid}.jpg (download URL)
            - https://bucket.s3.amazonaws.com/{user_id}/{uuid}?... (presigned URL)

    Returns:
        UUID string if found, None otherwise
    """
    match = _UUID_PATTERN.search(url)
    return match.group(1) if match else None



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import base64
import json
from pathlib import Path

from ...types import (
    Action,
    ActionEvent,
    ActionType,
    ImageEvent,
    LogEvent,
    ObserverEvent,
    PlanEvent,
    SplitEvent,
    StepEvent,
    parse_coords,
    parse_drag_coords,
    parse_scroll,
)


def _parse_action_coords(action: Action) -> dict | None:
    """Parse coordinates from action argument for cursor indicators.

    Returns:
        Dict with coordinates based on action type, or None if not applicable.
        - Click types: {"type": "click", "x": int, "y": int}
        - Drag: {"type": "drag", "x1": int, "y1": int, "x2": int, "y2": int}
        - Scroll: {"type": "scroll", "x": int, "y": int, "direction": str}
    """
    arg = action.argument.strip("()")

    match action.type:
        case (
            ActionType.CLICK
            | ActionType.LEFT_DOUBLE
            | ActionType.LEFT_TRIPLE
            | ActionType.RIGHT_SINGLE
        ):
            coords = parse_coords(arg)
            if coords:
                return {"type": "click", "x": coords[0], "y": coords[1]}
        case ActionType.DRAG:
            coords = parse_drag_coords(arg)
            if coords:
                return {
                    "type": "drag",
                    "x1": coords[0],
                    "y1": coords[1],
                    "x2": coords[2],
                    "y2": coords[3],
                }
        case ActionType.SCROLL:
            result = parse_scroll(arg)
            if result:
                return {
                    "type": "scroll",
                    "x": result[0],
                    "y": result[1],
                    "direction": result[2],
                }
    return None


def export_to_markdown(
    events: list[ObserverEvent],
    path: str,
    images_dir: str | None = None,
) -> None:
    """Export events to a Markdown file.

    Args:
        events: List of events to export.
        path: Path to the output Markdown file.
        images_dir: Directory to save images. If None, images are not saved.
    """
    output_path = Path(path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    if images_dir:
        images_path = Path(images_dir)
        images_path.mkdir(parents=True, exist_ok=True)

    lines: list[str] = ["# Agent Execution Report\n"]
    image_counter = 0

    for event in events:
        timestamp = event.timestamp.strftime("%H:%M:%S")

        match event:
            case StepEvent():
                lines.append(f"\n## Step {event.step_num}\n")
                lines.append(f"**Time:** {timestamp}\n")
                if event.task_id:
                    lines.append(f"**Task ID:** `{event.task_id}`\n")

                if isinstance(event.image, bytes):
                    if images_dir:
                        image_counter += 1
                        image_filename = f"step_{event.step_num}.png"
                        image_path = Path(images_dir) / image_filename
                        image_path.write_bytes(event.image)
                        rel_path = Path(images_dir).name / Path(image_filename)
                        lines.append(f"\n![Step {event.step_num}]({rel_path})\n")
                    else:
                        lines.append(
                            f"\n*[Screenshot captured - {len(event.image)} bytes]*\n"
                        )
                elif isinstance(event.image, str):
                    lines.append(f"\n**Screenshot URL:** {event.image}\n")

                if event.step.reason:
                    lines.append(f"\n**Reasoning:**\n> {event.step.reason}\n")

                if event.step.actions:
                    lines.append("\n**Planned Actions:**\n")
                    for action in event.step.actions:
                        count_str = (
                            f" (x{action.count})"
                            if action.count and action.count > 1
                            else ""
                        )
                        lines.append(
                            f"- `{action.type.value}`: {action.argument}{count_str}\n"
                        )

                if event.step.stop:
                    lines.append("\n**Status:** Task Complete\n")

            case ActionEvent():
                lines.append(f"\n### Actions Executed ({timestamp})\n")
                if event.error:
                    lines.append(f"\n**Error:** {event.error}\n")
                else:
                    lines.append("\n**Result:** Success\n")

            case LogEvent():
                lines.append(f"\n> **Log ({timestamp}):** {event.message}\n")

            case SplitEvent():
                if event.label:
                    lines.append(f"\n---\n\n### {event.label}\n")
                else:
                    lines.append("\n---\n")

            case ImageEvent():
                pass

            case PlanEvent():
                phase_titles = {
                    "initial": "Initial Planning",
                    "reflection": "Reflection",
                    "summary": "Summary",
                }
                phase_title = phase_titles.get(event.phase, event.phase.capitalize())
                lines.append(f"\n### {phase_title} ({timestamp})\n")
                if event.request_id:
                    lines.append(f"**Request ID:** `{event.request_id}`\n")

                if event.image:
                    if isinstance(event.image, bytes):
                        if images_dir:
                            image_counter += 1
                            image_filename = f"plan_{event.phase}_{image_counter}.png"
                            image_path = Path(images_dir) / image_filename
                            image_path.write_bytes(event.image)
                            rel_path = Path(images_dir).name / Path(image_filename)
                            lines.append(f"\n![{phase_title}]({rel_path})\n")
                        else:
                            lines.append(
                                f"\n*[Screenshot captured - {len(event.image)} bytes]*\n"
                            )
                    elif isinstance(event.image, str):
                        lines.append(f"\n**Screenshot URL:** {event.image}\n")

                if event.reasoning:
                    lines.append(f"\n**Reasoning:**\n> {event.reasoning}\n")

                if event.result:
                    lines.append(f"\n**Result:** {event.result}\n")

    output_path.write_text("".join(lines))


def _convert_events_for_html(events: list[ObserverEvent]) -> list[dict]:
    """Convert events to JSON-serializable format for HTML template."""
    result = []

    for event in events:
        timestamp = event.timestamp.strftime("%H:%M:%S")

        match event:
            case StepEvent():
                # Collect action coordinates for cursor indicators
                action_coords = []
                actions_list = []
                if event.step.actions:
                    for action in event.step.actions:
                        coords = _parse_action_coords(action)
                        if coords:
                            action_coords.append(coords)
                        actions_list.append(
                            {
                                "type": action.type.value,
                                "argument": action.argument,
                                "count": action.count or 1,
                            }
                        )

                # Handle image
                image_data = None
                if isinstance(event.image, bytes):
                    image_data = base64.b64encode(event.image).decode("utf-8")
                elif isinstance(event.image, str):
                    image_data = event.image

                result.append(
                    {
                        "event_type": "step",
                        "timestamp": timestamp,
                        "step_num": event.step_num,
                        "image": image_data,
                        "action_coords": action_coords,
                        "reason": event.step.reason,
                        "actions": actions_list,
                        "stop": event.step.stop,
                        "task_id": event.task_id,
                    }
                )

            case ActionEvent():
                result.append(
                    {
                        "event_type": "action",
                        "timestamp": timestamp,
                        "error": event.error,
                    }
                )

            case LogEvent():
                result.append(
                    {
                        "event_type": "log",
                        "timestamp": timestamp,
                        "message": event.message,
                    }
                )

            case SplitEvent():
                result.append(
                    {
                        "event_type": "split",
                        "timestamp": timestamp,
                        "label": event.label,
                    }
                )

            case ImageEvent():
                pass

            case PlanEvent():
                image_data = None
                if isinstance(event.image, bytes):
                    image_data = base64.b64encode(event.image).decode("utf-8")
                elif isinstance(event.image, str):
                    image_data = event.image

                result.append(
                    {
                        "event_type": "plan",
                        "timestamp": timestamp,
                        "phase": event.phase,
                        "image": image_data,
                        "reasoning": event.reasoning,
                        "result": event.result,
                        "request_id": event.request_id,
                    }
                )

    return result


def export_to_html(events: list[ObserverEvent], path: str) -> None:
    """Export events to a self-contained HTML file.

    Args:
        events: List of events to export.
        path: Path to the output HTML file.
    """
    output_path = Path(path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Load template
    template_path = Path(__file__).parent / "report_template.html"
    template = template_path.read_text()

    # Convert events to JSON
    events_data = _convert_events_for_html(events)
    events_json = json.dumps(events_data)

    # Replace placeholder
    html_content = template.replace("{EVENTS_DATA}", events_json)

    output_path.write_text(html_content)


def export_to_json(events: list[ObserverEvent], path: str) -> None:
    """Export events to a JSON file.

    Args:
        events: List of events to export.
        path: Path to the output JSON file.
    """
    output_path = Path(path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Convert events to JSON-serializable format
    json_events = []
    for event in events:
        # Handle bytes images before model_dump to avoid UTF-8 decode error
        if isinstance(event, (StepEvent, ImageEvent, PlanEvent)) and isinstance(
            getattr(event, "image", None), bytes
        ):
            # Dump without json mode first, then handle bytes manually
            event_dict = event.model_dump()
            event_dict["image"] = base64.b64encode(event.image).decode("utf-8")
            event_dict["image_encoding"] = "base64"
            # Convert datetime to string
            if "timestamp" in event_dict:
                event_dict["timestamp"] = event_dict["timestamp"].isoformat()
        else:
            event_dict = event.model_dump(mode="json")
        json_events.append(event_dict)

    output_path.write_text(json.dumps(json_events, indent=2, default=str))



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

# Re-export from types for convenience
from ...types import (
    ActionEvent,
    BaseEvent,
    ImageEvent,
    LogEvent,
    ObserverEvent,
    SplitEvent,
    StepEvent,
)

__all__ = [
    "ActionEvent",
    "BaseEvent",
    "ImageEvent",
    "LogEvent",
    "ObserverEvent",
    "SplitEvent",
    "StepEvent",
]



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------


import sys
from dataclasses import dataclass

from oagi.exceptions import check_optional_dependency

# Guard flag to prevent multiple DPI awareness calls on Windows
_dpi_awareness_set = False


@dataclass
class Screen:
    """
    Screen represents a single display screen.

    Attributes:
        name (str): The name of the screen.
        x (int): The x-coordinate of the top-left corner of the screen.
        y (int): The y-coordinate of the top-left corner of the screen.
        width (int): The width of the screen in pixels.
        height (int): The height of the screen in pixels.
        is_primary (bool): True if this is the primary screen, False otherwise.
    """

    name: str
    x: int
    y: int
    width: int
    height: int
    is_primary: bool = False


class ScreenManager:
    """
    ScreenManager is responsible for detecting and managing screens.
    """

    def __init__(self):
        self.screens = []
        # Enable DPI awareness if on Windows
        if sys.platform == "win32":
            self.enable_windows_dpi_awareness()

    def get_all_screens(self) -> list[Screen]:
        if self.screens:
            return self.screens
        if sys.platform == "darwin":
            screens = self._get_darwin_screen_info()
        elif sys.platform == "win32":
            screens = self._get_windows_screen_info()
        else:
            screens = self._get_linux_screen_info()
        # Find the primary screen
        primary_screen, alternative_screens = None, []
        for screen in screens:
            if screen.is_primary:
                primary_screen = screen
            else:
                alternative_screens.append(screen)
        # order the alternative_screens by x coordinate ascending and y coordinate ascending
        alternative_screens = sorted(
            alternative_screens, key=lambda item: (item.x, item.y)
        )
        # Add the primary screen to the front of the list if it exists
        if primary_screen:
            self.screens = [primary_screen]
        self.screens += alternative_screens
        return self.screens

    def _get_darwin_screen_info(self) -> list[Screen]:
        """
        Get screen information for macOS using AppKit.

        Returns:
            list[Screen]: A list of Screen objects representing all detected screens.
        """
        check_optional_dependency("AppKit", "ScreenManager", "desktop")
        import AppKit  # noqa: PLC0415

        # Force the RunLoop to update once
        # This "accepts input" which forces macOS to update screen geometry
        loop = AppKit.NSRunLoop.currentRunLoop()
        loop.acceptInputForMode_beforeDate_(
            AppKit.NSDefaultRunLoopMode, AppKit.NSDate.distantPast()
        )
        # Retrieve screen information using AppKit
        screens = AppKit.NSScreen.screens()
        screen_list = []
        for screen in screens:
            frame = screen.frame()
            # Origin (0,0) is bottom-left of the primary screen
            x, y = int(frame.origin.x), int(frame.origin.y)
            width, height = int(frame.size.width), int(frame.size.height)
            name = screen.localizedName()
            # Normalize the origin to Top-Left
            y = int(AppKit.NSScreen.screens()[0].frame().size.height) - (y + height)
            screen_list.append(Screen(name, x, y, width, height, x == 0 and y == 0))
        return screen_list

    def _get_windows_screen_info(self) -> list[Screen]:
        """
        Get screen information for Windows using mss.

        Returns:
            list[Screen]: A list of Screen objects representing all detected screens.
        """
        check_optional_dependency("mss", "ScreenManager", "desktop")
        import mss  # noqa: PLC0415

        screen_list = []
        for index, screen in enumerate(mss.mss().monitors[1:]):
            screen_list.append(
                Screen(
                    f"DISPLAY{index}",
                    screen["left"],
                    screen["top"],
                    screen["width"],
                    screen["height"],
                    screen["top"] == 0 and screen["left"] == 0,
                )
            )
        return screen_list

    def _get_linux_screen_info(self) -> list[Screen]:
        """
        Get screen information for Linux and other platforms as default.

        Returns:
            list[Screen]: A list of Screen objects representing all detected screens.
        """
        check_optional_dependency("screeninfo", "ScreenManager", "desktop")
        import screeninfo  # noqa: PLC0415

        screen_list = []
        for screen in screeninfo.get_monitors():
            screen_list.append(
                Screen(
                    screen.name,
                    screen.x,
                    screen.y,
                    screen.width,
                    screen.height,
                    screen.is_primary,
                )
            )
        return screen_list

    @staticmethod
    def enable_windows_dpi_awareness():
        """
        Enable per-monitor DPI awareness to fix multi-monitor scaling issues.

        On Windows with mixed scaling between monitors, applications that are not
        DPI-aware will have their coordinates virtualized, causing clicks/moves to
        land at incorrect positions. Enabling DPI awareness ensures PyAutoGUI and mss
        works in physical pixels across all monitors.

        This method is idempotent - subsequent calls after the first successful call
        will be no-ops.
        """
        global _dpi_awareness_set
        if _dpi_awareness_set:
            return

        import ctypes  # noqa: PLC0415

        try:
            # For Windows 8.1 and Windows 10/11
            # 2 = PROCESS_PER_MONITOR_DPI_AWARE
            PROCESS_PER_MONITOR_DPI_AWARE = 2
            ctypes.windll.shcore.SetProcessDpiAwareness(PROCESS_PER_MONITOR_DPI_AWARE)
            _dpi_awareness_set = True
        except Exception:
            try:
                # Fallback for older Windows versions
                ctypes.windll.user32.SetProcessDPIAware()
                _dpi_awareness_set = True
            except Exception:
                raise RuntimeError("Could not set DPI awareness")



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import pytest

from oagi.types import ActionType, Step
from oagi.utils.output_parser import _parse_action, _split_actions, parse_raw_output


class TestParseRawOutput:
    """Test parse_raw_output function."""

    def test_parse_single_click_action(self):
        raw = "<|think_start|>Click the button<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>"
        step = parse_raw_output(raw)

        assert step.reason == "Click the button"
        assert len(step.actions) == 1
        assert step.actions[0].type == ActionType.CLICK
        assert step.actions[0].argument == "300, 150"
        assert step.stop is False

    def test_parse_finish_action_sets_stop(self):
        raw = "<|think_start|>Task complete<|think_end|>\n<|action_start|>finish()<|action_end|>"
        step = parse_raw_output(raw)

        assert step.reason == "Task complete"
        assert len(step.actions) == 1
        assert step.actions[0].type == ActionType.FINISH
        assert step.stop is True

    def test_parse_multiple_actions_with_ampersand(self):
        raw = "<|think_start|>Do two things<|think_end|>\n<|action_start|>click(100, 200) & type(hello)<|action_end|>"
        step = parse_raw_output(raw)

        assert step.reason == "Do two things"
        assert len(step.actions) == 2
        assert step.actions[0].type == ActionType.CLICK
        assert step.actions[1].type == ActionType.TYPE

    def test_parse_empty_output(self):
        step = parse_raw_output("")

        assert step.reason == ""
        assert step.actions == []
        assert step.stop is False

    def test_parse_missing_think_tags(self):
        raw = "<|action_start|>click(100, 200)<|action_end|>"
        step = parse_raw_output(raw)

        assert step.reason == ""
        assert len(step.actions) == 1
        assert step.actions[0].type == ActionType.CLICK

    def test_parse_missing_action_tags(self):
        raw = "<|think_start|>Just thinking<|think_end|>"
        step = parse_raw_output(raw)

        assert step.reason == "Just thinking"
        assert step.actions == []
        assert step.stop is False

    def test_parse_multiline_reason(self):
        raw = "<|think_start|>First line\nSecond line\nThird line<|think_end|>\n<|action_start|>click(100, 200)<|action_end|>"
        step = parse_raw_output(raw)

        assert "First line" in step.reason
        assert "Second line" in step.reason
        assert "Third line" in step.reason

    def test_parse_all_action_types(self):
        raw = """<|think_start|>Test all actions<|think_end|>
<|action_start|>click(100, 200) & left_double(200, 300) & left_triple(300, 400) & right_single(400, 500) & drag(100, 100, 500, 500) & hotkey(ctrl+c) & type(hello) & scroll(500, 300, up) & wait() & finish()<|action_end|>"""
        step = parse_raw_output(raw)

        assert len(step.actions) == 10
        assert step.actions[0].type == ActionType.CLICK
        assert step.actions[1].type == ActionType.LEFT_DOUBLE
        assert step.actions[2].type == ActionType.LEFT_TRIPLE
        assert step.actions[3].type == ActionType.RIGHT_SINGLE
        assert step.actions[4].type == ActionType.DRAG
        assert step.actions[5].type == ActionType.HOTKEY
        assert step.actions[6].type == ActionType.TYPE
        assert step.actions[7].type == ActionType.SCROLL
        assert step.actions[8].type == ActionType.WAIT
        assert step.actions[9].type == ActionType.FINISH
        assert step.stop is True

    def test_parse_invalid_action_skipped(self):
        raw = "<|think_start|>Test<|think_end|>\n<|action_start|>click(100, 200) & invalid_action() & type(hello)<|action_end|>"
        step = parse_raw_output(raw)

        assert len(step.actions) == 2
        assert step.actions[0].type == ActionType.CLICK
        assert step.actions[1].type == ActionType.TYPE


class TestSplitActions:
    """Test _split_actions function."""

    def test_split_single_action(self):
        result = _split_actions("click(100, 200)")
        assert result == ["click(100, 200)"]

    def test_split_multiple_actions(self):
        result = _split_actions("click(100, 200) & type(hello)")
        assert result == ["click(100, 200)", "type(hello)"]

    def test_split_preserves_nested_parentheses(self):
        result = _split_actions("type(func(a, b)) & click(100, 200)")
        assert result == ["type(func(a, b))", "click(100, 200)"]

    def test_split_empty_string(self):
        result = _split_actions("")
        assert result == []

    def test_split_whitespace_handling(self):
        result = _split_actions("  click(100, 200)   &   type(hello)  ")
        assert result == ["click(100, 200)", "type(hello)"]

    def test_split_no_ampersand(self):
        result = _split_actions("click(100, 200)")
        assert result == ["click(100, 200)"]

    def test_split_three_actions(self):
        result = _split_actions("click(100, 200) & type(hello) & finish()")
        assert result == ["click(100, 200)", "type(hello)", "finish()"]


class TestParseAction:
    """Test _parse_action function."""

    @pytest.mark.parametrize(
        "action_text,expected_type,expected_arg",
        [
            ("click(100, 200)", ActionType.CLICK, "100, 200"),
            ("left_double(100, 200)", ActionType.LEFT_DOUBLE, "100, 200"),
            ("left_triple(100, 200)", ActionType.LEFT_TRIPLE, "100, 200"),
            ("right_single(100, 200)", ActionType.RIGHT_SINGLE, "100, 200"),
            ("drag(100, 100, 500, 500)", ActionType.DRAG, "100, 100, 500, 500"),
            ("type(hello world)", ActionType.TYPE, "hello world"),
            ("wait()", ActionType.WAIT, ""),
            ("finish()", ActionType.FINISH, ""),
        ],
    )
    def test_parse_basic_actions(self, action_text, expected_type, expected_arg):
        action = _parse_action(action_text)
        assert action is not None
        assert action.type == expected_type
        assert action.argument == expected_arg
        assert action.count == 1

    def test_parse_hotkey_with_count(self):
        action = _parse_action("hotkey(ctrl+c, 3)")
        assert action is not None
        assert action.type == ActionType.HOTKEY
        assert action.argument == "ctrl+c"
        assert action.count == 3

    def test_parse_hotkey_without_count(self):
        action = _parse_action("hotkey(ctrl+c)")
        assert action is not None
        assert action.type == ActionType.HOTKEY
        assert action.argument == "ctrl+c"
        assert action.count == 1

    def test_parse_scroll_with_count(self):
        action = _parse_action("scroll(500, 300, up, 5)")
        assert action is not None
        assert action.type == ActionType.SCROLL
        assert action.argument == "500,300,up"
        assert action.count == 5

    def test_parse_scroll_without_count(self):
        action = _parse_action("scroll(500, 300, down)")
        assert action is not None
        assert action.type == ActionType.SCROLL
        assert action.argument == "500, 300, down"
        assert action.count == 1

    def test_parse_invalid_action_type(self):
        action = _parse_action("invalid_action(123)")
        assert action is None

    def test_parse_malformed_syntax_no_parens(self):
        action = _parse_action("click 100, 200")
        assert action is None

    def test_parse_empty_string(self):
        action = _parse_action("")
        assert action is None

    def test_parse_uppercase_action(self):
        action = _parse_action("CLICK(100, 200)")
        assert action is not None
        assert action.type == ActionType.CLICK

    def test_parse_scroll_invalid_count_defaults_to_one(self):
        action = _parse_action("scroll(500, 300, up, abc)")
        assert action is not None
        assert action.count == 1

    def test_parse_hotkey_invalid_count_defaults_to_one(self):
        action = _parse_action("hotkey(ctrl+c, abc)")
        assert action is not None
        assert action.count == 1


class TestEdgeCases:
    """Test edge cases and robustness."""

    def test_parse_extra_whitespace(self):
        raw = "  <|think_start|>  thinking  <|think_end|>  \n  <|action_start|>  click(100, 200)  <|action_end|>  "
        step = parse_raw_output(raw)

        assert step.reason == "thinking"
        assert len(step.actions) == 1

    def test_parse_type_with_special_characters(self):
        raw = "<|think_start|>Type special<|think_end|>\n<|action_start|>type(!@#$%^*())<|action_end|>"
        step = parse_raw_output(raw)

        assert len(step.actions) == 1
        assert step.actions[0].argument == "!@#$%^*()"

    def test_parse_type_with_spaces(self):
        raw = "<|think_start|>Type text<|think_end|>\n<|action_start|>type(hello world 123)<|action_end|>"
        step = parse_raw_output(raw)

        assert step.actions[0].argument == "hello world 123"

    def test_returns_step_type(self):
        raw = "<|think_start|>Test<|think_end|>\n<|action_start|>click(100, 200)<|action_end|>"
        step = parse_raw_output(raw)

        assert isinstance(step, Step)

    def test_multiple_finish_actions_stop_true(self):
        raw = "<|think_start|>Test<|think_end|>\n<|action_start|>finish() & finish()<|action_end|>"
        step = parse_raw_output(raw)

        assert step.stop is True
        assert len(step.actions) == 2

    def test_finish_with_other_actions(self):
        raw = "<|think_start|>Test<|think_end|>\n<|action_start|>click(100, 200) & finish()<|action_end|>"
        step = parse_raw_output(raw)

        assert step.stop is True
        assert len(step.actions) == 2



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import sys
from unittest.mock import MagicMock, patch

import pytest

if sys.platform != "darwin":
    pytest.skip(
        "Skipping macOS-specific tests on non-macOS platform", allow_module_level=True
    )

from oagi.handler import _macos  # noqa: E402
from oagi.handler.pyautogui_action_handler import (
    PyautoguiActionHandler,
    PyautoguiConfig,
)
from oagi.types import Action, ActionType


@pytest.fixture
def mock_pyautogui():
    with patch("oagi.handler.pyautogui_action_handler.pyautogui") as mock:
        mock.size.return_value = (1920, 1080)
        yield mock


@pytest.fixture
def mock_quartz():
    mock = MagicMock()
    # Set up constants
    mock.kCGEventLeftMouseDown = 1
    mock.kCGEventLeftMouseUp = 2
    mock.kCGMouseButtonLeft = 0
    mock.kCGMouseEventClickState = 99
    mock.kCGHIDEventTap = 55
    return mock


def test_macos_click_implementation(mock_quartz):
    """Test the low-level macos_click function."""
    with (
        patch.object(_macos, "Quartz", mock_quartz),
        patch("oagi.handler._macos.pyautogui") as mock_pg,
    ):
        # Test triple click
        _macos.macos_click(100, 200, clicks=3)

        # Should move mouse
        mock_pg.moveTo.assert_called_once_with(100, 200)

        # Should create 6 events (3 clicks * 2 events/click)
        assert mock_quartz.CGEventCreateMouseEvent.call_count == 6

        # Should post 6 events
        assert mock_quartz.CGEventPost.call_count == 6

        # Verify click states
        # We expect setIntegerValue to be called 6 times
        calls = mock_quartz.CGEventSetIntegerValueField.call_args_list
        assert len(calls) == 6

        # Check pairs of calls have increasing click states
        # Click 1
        assert calls[0][0][2] == 1
        assert calls[1][0][2] == 1
        # Click 2
        assert calls[2][0][2] == 2
        assert calls[3][0][2] == 2
        # Click 3
        assert calls[4][0][2] == 3
        assert calls[5][0][2] == 3


def test_handler_calls_macos_click_double(mock_pyautogui):
    """Test that handler calls macos_click for double click on macOS."""
    handler = PyautoguiActionHandler(config=PyautoguiConfig(post_batch_delay=0))
    action = Action(type=ActionType.LEFT_DOUBLE, argument="500, 500", count=1)

    with (
        patch.object(sys, "platform", "darwin"),
        patch("oagi.handler.pyautogui_action_handler._macos") as mock_macos_module,
    ):
        handler([action])

        mock_macos_module.macos_click.assert_called_once_with(960, 540, clicks=2)
        mock_pyautogui.doubleClick.assert_not_called()


def test_handler_calls_macos_click_triple(mock_pyautogui):
    """Test that handler calls macos_click for triple click on macOS."""
    handler = PyautoguiActionHandler(config=PyautoguiConfig(post_batch_delay=0))
    action = Action(type=ActionType.LEFT_TRIPLE, argument="500, 500", count=1)

    with (
        patch.object(sys, "platform", "darwin"),
        patch("oagi.handler.pyautogui_action_handler._macos") as mock_macos_module,
    ):
        handler([action])

        mock_macos_module.macos_click.assert_called_once_with(960, 540, clicks=3)
        mock_pyautogui.tripleClick.assert_not_called()



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import re

from ..types.models.action import Action, ActionType
from ..types.models.step import Step


def parse_raw_output(raw_output: str) -> Step:
    """Parse raw LLM output into structured Step format.

    Expected format:
    <|think_start|> reasoning text <|think_end|>
    <|action_start|> action1(args) & action2(args) & ... <|action_end|>

    Args:
        raw_output: Raw text output from the LLM

    Returns:
        Step object with parsed reasoning and actions
    """
    # Extract reasoning/thinking
    think_pattern = r"<\|think_start\|>(.*?)<\|think_end\|>"
    think_match = re.search(think_pattern, raw_output, re.DOTALL)
    reason = think_match.group(1).strip() if think_match else ""

    # Extract action block
    action_pattern = r"<\|action_start\|>(.*?)<\|action_end\|>"
    action_match = re.search(action_pattern, raw_output, re.DOTALL)

    actions: list[Action] = []
    stop = False

    if action_match:
        action_block = action_match.group(1).strip()
        action_texts = _split_actions(action_block)

        for action_text in action_texts:
            parsed_action = _parse_action(action_text.strip())
            if parsed_action:
                actions.append(parsed_action)
                if parsed_action.type == ActionType.FINISH:
                    stop = True

    return Step(reason=reason, actions=actions, stop=stop)


def _split_actions(action_block: str) -> list[str]:
    """Split action block by & separator, but only when & is outside parentheses.

    Note: This parser does NOT handle '&' inside quoted strings.
    E.g., type("a&b") would incorrectly split. The LLM should avoid
    this pattern by using alternative escape sequences.

    Args:
        action_block: String containing one or more actions separated by &

    Returns:
        List of individual action strings
    """
    actions: list[str] = []
    current_action: list[str] = []
    paren_level = 0

    for char in action_block:
        if char == "(":
            paren_level += 1
            current_action.append(char)
        elif char == ")":
            paren_level -= 1
            current_action.append(char)
        elif char == "&" and paren_level == 0:
            action_str = "".join(current_action).strip()
            if action_str:
                actions.append(action_str)
            current_action = []
        else:
            current_action.append(char)

    # Add the last action
    action_str = "".join(current_action).strip()
    if action_str:
        actions.append(action_str)

    return actions


def _parse_action(action_text: str) -> Action | None:
    """Parse individual action text into Action object.

    Expected formats:
    - click(x, y) # left-click at position
    - left_double(x, y) # left-double-click at position
    - left_triple(x, y) # left-triple-click at position
    - right_single(x, y) # right-click at position
    - drag(x1, y1, x2, y2) # drag from (x1, y1) to (x2, y2)
    - hotkey(key, c) # press key c times
    - type(text) # type text string
    - scroll(x, y, direction, c) # scroll at position
    - wait() # wait for a while
    - finish() # indicate task is finished

    Args:
        action_text: String representation of a single action

    Returns:
        Action object or None if parsing fails
    """
    # Match action format: action_type(arguments)
    match = re.match(r"(\w+)\((.*)\)", action_text.strip())
    if not match:
        return None

    action_type = match.group(1).lower()
    arguments = match.group(2).strip()

    # Parse count from arguments for actions that support it
    count = 1

    # Validate and map action type to enum
    try:
        action_enum = ActionType(action_type)
    except ValueError:
        return None

    # Parse specific action types and extract count where applicable
    match action_enum:
        case ActionType.HOTKEY:
            # hotkey(key, c) - press key c times
            args = arguments.rsplit(",", 1)
            if len(args) >= 2 and args[1].strip():
                key = args[0].strip()
                try:
                    count = int(args[1].strip())
                except ValueError:
                    count = 1
            else:
                key = arguments.strip()
                count = 1
            arguments = key

        case ActionType.SCROLL:
            # scroll(x, y, direction, c) - scroll at position
            args = arguments.split(",")
            if len(args) >= 4:
                x = args[0].strip()
                y = args[1].strip()
                direction = args[2].strip()
                try:
                    count = int(args[3].strip())
                except (ValueError, IndexError):
                    count = 1
                # Reconstruct arguments without count
                arguments = f"{x},{y},{direction}"

        case _:
            # For other actions, use default count of 1
            pass

    return Action(type=action_enum, argument=arguments, count=count)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------
"""Action converters for VLM support.

This module provides the base class and OAGI implementation for action converters.
Third parties can inherit from BaseActionConverter to create custom converters.

Example usage:
    from oagi.converters import OagiActionConverter, ConverterConfig

    # Configure for 1920x1080 sandbox
    config = ConverterConfig(sandbox_width=1920, sandbox_height=1080)
    converter = OagiActionConverter(config=config)

    # Convert OAGI actions to pyautogui strings
    result = converter(actions)  # list[str]

    # Convert to runtime API steps
    for cmd in result:
        step = converter.action_string_to_step(cmd)
        # Execute step via runtime API...

Creating custom converters:
    from oagi.converters import BaseActionConverter, ConverterConfig

    class MyActionConverter(BaseActionConverter[MyAction]):
        @property
        def coord_width(self) -> int:
            return 1000  # Your model's coordinate width

        @property
        def coord_height(self) -> int:
            return 1000  # Your model's coordinate height

        def _convert_single_action(self, action: MyAction) -> list[str]:
            # Convert action to pyautogui command strings
            ...

        def serialize_actions(self, actions: list[MyAction]) -> list[dict]:
            # Serialize actions for trajectory logging
            ...
"""

from .base import BaseActionConverter, ConverterConfig
from .oagi import OagiActionConverter

__all__ = [
    "BaseActionConverter",
    "ConverterConfig",
    "OagiActionConverter",
]



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import time

from pydantic import BaseModel, Field

from oagi.handler.screen_manager import Screen

from ..constants import DEFAULT_STEP_DELAY
from ..types import Action, ActionType, parse_coords, parse_drag_coords, parse_scroll
from .capslock_manager import CapsLockManager
from .utils import CoordinateScaler, normalize_key, parse_hotkey
from .wayland_support import Ydotool, get_screen_size


class YdotoolConfig(BaseModel):
    """Configuration for YdotoolActionHandler."""

    scroll_amount: int = Field(
        default=20,
        description="Amount to scroll (positive for up, negative for down)",
    )
    wait_duration: float = Field(
        default=1.0, description="Duration for wait actions in seconds"
    )
    action_pause: float = Field(
        default=0.5, description="Pause between Ydotool actions in seconds"
    )
    capslock_mode: str = Field(
        default="session",
        description="Caps lock handling mode: 'session' (internal state) or 'system' (OS-level)",
    )
    socket_address: str = Field(
        default="",
        description="Custom socket address for ydotool (e.g., '/tmp/ydotool.sock')",
    )
    post_batch_delay: float = Field(
        default=DEFAULT_STEP_DELAY,
        ge=0,
        description="Delay after executing all actions in a batch (seconds). "
        "Allows UI to settle before next screenshot.",
    )


class YdotoolActionHandler(Ydotool):
    def __init__(self, config: YdotoolConfig | None = None) -> None:
        """
        Handles actions to be executed using Ydotool.

        This class provides functionality for handling and executing a sequence of
        actions using the Ydotool. It processes a list of actions and executes
        them as per the implementation.

        Methods:
            __call__: Executes the provided list of actions.

        Args:
            actions (list[Action]): List of actions to be processed and executed.
        """
        # Use default config if none provided
        self.config = config or YdotoolConfig()
        super().__init__(socket_address=self.config.socket_address)
        # Get screen dimensions for coordinate denormalization
        self.screen_width, self.screen_height = get_screen_size()
        # Set default delay between actions
        self.action_pause = self.config.action_pause
        # Initialize caps lock manager
        self.caps_manager = CapsLockManager(mode=self.config.capslock_mode)
        # The origin position of coordinates (the top-left corner of the screen)
        self.origin_x, self.origin_y = 0, 0
        # Initialize coordinate scaler
        self._coord_scaler = CoordinateScaler(
            source_width=1000,
            source_height=1000,
            target_width=self.screen_width,
            target_height=self.screen_height,
        )

    def reset(self):
        """Reset handler state.

        Called at automation start/end and when FINISH action is received.
        Resets the internal capslock state.
        """
        self.caps_manager.reset()

    def set_target_screen(self, screen: Screen) -> None:
        """Set the target screen for the action handler.

        Args:
            screen (Screen): The screen object to set as the target.
        """
        self.screen_width, self.screen_height = screen.width, screen.height
        self.origin_x, self.origin_y = screen.x, screen.y
        self._coord_scaler = CoordinateScaler(
            source_width=1000,
            source_height=1000,
            target_width=self.screen_width,
            target_height=self.screen_height,
        )

    def _execute_action(self, action: Action) -> bool:
        """
        Execute a group of actions and return whether FINISH is reached.
        """
        finished = False
        arg = (action.argument or "").strip()
        count = int(action.count or 1)

        match action.type:
            case ActionType.DRAG:
                x1, y1, x2, y2 = self._parse_drag_coords(arg)
                self.drag(x1, y1, x2, y2, count=count)

            case ActionType.SCROLL:
                x, y, direction = self._parse_scroll(arg)
                scroll_amount = (
                    self.config.scroll_amount
                    if direction == "up"
                    else -self.config.scroll_amount
                )
                self.scroll(scroll_amount)

            case ActionType.RIGHT_SINGLE:
                x, y = self._parse_coords(arg)
                self.click(x, y, right=True, count=count)

            case ActionType.LEFT_DOUBLE:
                x, y = self._parse_coords(arg)
                self.click(x, y, count=2 * count)

            case ActionType.LEFT_TRIPLE:
                x, y = self._parse_coords(arg)
                self.click(x, y, count=3 * count)

            case ActionType.CLICK:
                x, y = self._parse_coords(arg)
                self.click(x, y, count=count)

            case ActionType.HOTKEY:
                keys = self._parse_hotkey(arg)
                # Check if this is a caps lock key press
                if len(keys) == 1 and keys[0] == "capslock":
                    if self.caps_manager.should_use_system_capslock():
                        # System mode: use OS-level caps lock
                        self.hotkey(["capslock"])
                    else:
                        # Session mode: toggle internal state
                        self.caps_manager.toggle()
                else:
                    # Regular hotkey combination
                    self.hotkey(keys, count=count)

            case ActionType.TYPE:
                # Remove quotes if present
                text = arg.strip("\"'")
                # Apply caps lock transformation if needed
                text = self.caps_manager.transform_text(text)
                self._run_ydotool(["type", text], count=count)

            case ActionType.FINISH:
                # Task completion - reset handler state
                self.reset()

            case ActionType.WAIT:
                # Wait for a short period
                time.sleep(self.config.wait_duration)

            case ActionType.CALL_USER:
                # Call user - implementation depends on requirements
                print("User intervention requested")

            case _:
                print(f"Unknown action type: {action.type}")

        return finished

    def _denormalize_coords(self, x: float, y: float) -> tuple[int, int]:
        """Convert coordinates from 0-1000 range to actual screen coordinates."""
        screen_x, screen_y = self._coord_scaler.scale(x, y, prevent_failsafe=True)
        # Add origin offset for multi-screen support
        return screen_x + self.origin_x, screen_y + self.origin_y

    def _normalize_key(self, key: str) -> str:
        """Normalize key names for consistency."""
        return normalize_key(key)

    def _parse_coords(self, args_str: str) -> tuple[int, int]:
        """Extract x, y coordinates from argument string."""
        coords = parse_coords(args_str)
        if not coords:
            raise ValueError(f"Invalid coordinates format: {args_str}")
        return self._denormalize_coords(coords[0], coords[1])

    def _parse_drag_coords(self, args_str: str) -> tuple[int, int, int, int]:
        """Extract x1, y1, x2, y2 coordinates from drag argument string."""
        coords = parse_drag_coords(args_str)
        if not coords:
            raise ValueError(f"Invalid drag coordinates format: {args_str}")
        x1, y1 = self._denormalize_coords(coords[0], coords[1])
        x2, y2 = self._denormalize_coords(coords[2], coords[3])
        return x1, y1, x2, y2

    def _parse_scroll(self, args_str: str) -> tuple[int, int, str]:
        """Extract x, y, direction from scroll argument string."""
        result = parse_scroll(args_str)
        if not result:
            raise ValueError(f"Invalid scroll format: {args_str}")
        x, y = self._denormalize_coords(result[0], result[1])
        return x, y, result[2]

    def _parse_hotkey(self, args_str: str) -> list[str]:
        """Parse hotkey string into list of keys."""
        return parse_hotkey(args_str.strip("()"), validate=False)

    def __call__(self, actions: list[Action]) -> None:
        """Execute the provided list of actions."""
        for action in actions:
            try:
                self._execute_action(action)
            except Exception as e:
                print(f"Error executing action {action.type}: {e}")
                raise

        # Wait after batch for UI to settle before next screenshot
        if self.config.post_batch_delay > 0:
            time.sleep(self.config.post_batch_delay)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

from enum import Enum

from ...types import LogEvent, ObserverEvent, SplitEvent
from .exporters import export_to_html, export_to_json, export_to_markdown


class ExportFormat(str, Enum):
    """Supported export formats."""

    MARKDOWN = "markdown"
    HTML = "html"
    JSON = "json"


class AsyncAgentObserver:
    """Records agent execution events and exports to various formats.

    This class implements the AsyncObserver protocol and provides
    functionality for recording events during agent execution and
    exporting them to Markdown or HTML formats.
    """

    def __init__(self) -> None:
        self.events: list[ObserverEvent] = []

    async def on_event(self, event: ObserverEvent) -> None:
        """Record an event.

        Args:
            event: The event to record.
        """
        self.events.append(event)

    def add_log(self, message: str) -> None:
        """Add a custom log message.

        Args:
            message: The log message to add.
        """
        self.events.append(LogEvent(message=message))

    def add_split(self, label: str = "") -> None:
        """Add a visual separator.

        Args:
            label: Optional label for the separator.
        """
        self.events.append(SplitEvent(label=label))

    def clear(self) -> None:
        """Clear all recorded events."""
        self.events.clear()

    def get_events_by_step(self, step_num: int) -> list[ObserverEvent]:
        """Get all events for a specific step.

        Args:
            step_num: The step number to filter by.

        Returns:
            List of events for the specified step.
        """
        return [
            event
            for event in self.events
            if hasattr(event, "step_num") and event.step_num == step_num
        ]

    def export(
        self,
        format: ExportFormat | str,
        path: str,
        images_dir: str | None = None,
    ) -> None:
        """Export recorded events to a file.

        Args:
            format: Export format (markdown, html, json)
            path: Path to the output file.
            images_dir: Directory to save images (markdown only).
        """
        if isinstance(format, str):
            format = ExportFormat(format.lower())

        match format:
            case ExportFormat.MARKDOWN:
                export_to_markdown(self.events, path, images_dir)
            case ExportFormat.HTML:
                export_to_html(self.events, path)
            case ExportFormat.JSON:
                export_to_json(self.events, path)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

from unittest.mock import patch

import pytest

from oagi.actor import Actor
from oagi.constants import MODEL_ACTOR
from oagi.types import ActionType, Step
from oagi.types.models import UploadFileResponse


@pytest.fixture
def actor(mock_sync_client):
    """Create an Actor instance with mocked client."""
    return Actor(api_key="test-key", base_url="https://test.example.com")


@pytest.fixture
def mock_upload_file_response():
    """Mock UploadFileResponse for S3 upload."""
    return UploadFileResponse(
        url="https://s3.amazonaws.com/presigned-url",
        uuid="test-uuid-123",
        expires_at=1677652888,
        file_expires_at=1677739288,
        download_url="https://cdn.example.com/test-uuid-123",
    )


class TestActorInit:
    def test_init_with_parameters(self, mock_sync_client):
        actor = Actor(api_key="test-key", base_url="https://test.example.com")
        assert actor.api_key == "test-key"
        assert actor.base_url == "https://test.example.com"
        assert actor.task_id is not None
        assert isinstance(actor.task_id, str)
        assert len(actor.task_id) == 32  # UUID hex without dashes
        assert actor.task_description is None
        assert actor.model == MODEL_ACTOR
        assert actor.message_history == []

    def test_init_with_custom_model(self, mock_sync_client):
        actor = Actor(
            api_key="test-key",
            base_url="https://test.example.com",
            model="custom-model",
        )
        assert actor.model == "custom-model"

    def test_init_with_env_vars(self, mock_sync_client):
        with patch.dict(
            "os.environ",
            {"OAGI_BASE_URL": "https://env.example.com", "OAGI_API_KEY": "env-key"},
        ):
            mock_sync_client.api_key = "env-key"
            mock_sync_client.base_url = "https://env.example.com"
            actor = Actor()
            assert actor.api_key == "env-key"
            assert actor.base_url == "https://env.example.com"

    def test_init_parameters_override_env_vars(self, mock_sync_client):
        with patch.dict(
            "os.environ",
            {"OAGI_BASE_URL": "https://env.example.com", "OAGI_API_KEY": "env-key"},
        ):
            actor = Actor(
                api_key="override-key", base_url="https://override.example.com"
            )
            assert actor.api_key == "test-key"  # From mock_sync_client
            assert actor.base_url == "https://test.example.com"  # From mock_sync_client


class TestActorInitTask:
    def test_init_task_success(self, actor):
        original_task_id = actor.task_id

        actor.init_task("Test task description", max_steps=10)

        assert actor.task_description == "Test task description"
        # task_id is regenerated on init_task to create a fresh task
        assert actor.task_id != original_task_id
        assert isinstance(actor.task_id, str)
        assert len(actor.task_id) == 32  # UUID hex without dashes


class TestActorStep:
    def test_step_with_image_object(
        self,
        actor,
        mock_image,
        sample_step,
        sample_usage_obj,
        mock_upload_file_response,
    ):
        actor.task_description = "Test task"
        actor.task_id = "existing-task"

        # Setup mocks
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        result = actor.step(mock_image)

        # Verify Image.read() was called
        mock_image.read.assert_called_once()

        # Verify S3 upload was called
        actor.client.put_s3_presigned_url.assert_called_once()

        # Verify chat_completion was called with messages
        actor.client.chat_completion.assert_called_once()
        call_args = actor.client.chat_completion.call_args
        assert call_args[1]["model"] == MODEL_ACTOR
        assert "messages" in call_args[1]
        assert call_args[1]["temperature"] is None

        # Verify returned Step
        assert isinstance(result, Step)
        assert result.stop is False

        # Verify message_history was updated (user + assistant)
        assert len(actor.message_history) == 2
        assert actor.message_history[0]["role"] == "user"
        assert actor.message_history[1]["role"] == "assistant"

    def test_step_with_bytes_directly(
        self, actor, sample_step, sample_usage_obj, mock_upload_file_response
    ):
        actor.task_description = "Test task"
        original_task_id = actor.task_id

        # Setup mocks
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        image_bytes = b"raw image bytes"
        result = actor.step(image_bytes)

        # Verify S3 upload was called
        actor.client.put_s3_presigned_url.assert_called_once_with(image_bytes)

        # Verify chat_completion was called with messages
        call_args = actor.client.chat_completion.call_args
        assert call_args[1]["model"] == MODEL_ACTOR
        assert "messages" in call_args[1]
        assert call_args[1]["temperature"] is None

        # task_id doesn't change
        assert actor.task_id == original_task_id

        # Verify returned Step
        assert isinstance(result, Step)
        assert result.stop is False

    def test_step_with_url_directly(self, actor, sample_step, sample_usage_obj):
        """Test that step with URL skips S3 upload."""
        actor.task_description = "Test task"

        # Setup mocks
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        screenshot_url = "https://cdn.example.com/screenshot.png"
        result = actor.step(screenshot_url)

        # Verify S3 upload was NOT called (URL used directly)
        actor.client.put_s3_presigned_url.assert_not_called()

        # Verify chat_completion was called with messages containing the URL
        actor.client.chat_completion.assert_called_once()
        # After step: message_history has 2 (user + assistant)
        assert len(actor.message_history) == 2
        assert screenshot_url in str(actor.message_history[0])  # URL in user message

        assert isinstance(result, Step)

    def test_step_without_init_task_raises_error(self, actor):
        with pytest.raises(
            ValueError, match="Task description must be set. Call init_task\\(\\) first"
        ):
            actor.step(b"image bytes")

    def test_step_with_completed_response(
        self, actor, completed_step, sample_usage_obj, mock_upload_file_response
    ):
        actor.task_description = "Test task"
        actor.task_id = "task-456"

        # Setup mocks
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            completed_step,
            "<|think_start|>done<|think_end|>\n<|action_start|>finish()<|action_end|>",
            sample_usage_obj,
        )

        result = actor.step(b"image bytes")

        assert result.stop is True
        assert result.reason == "The task has been completed successfully"
        assert len(result.actions) == 1
        assert result.actions[0].type == ActionType.FINISH

    def test_step_handles_exception(self, actor, mock_upload_file_response):
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.side_effect = Exception("API Error")

        with pytest.raises(Exception, match="API Error"):
            actor.step(b"image bytes")

    def test_step_raises_error_when_max_steps_reached(
        self, actor, sample_step, sample_usage_obj, mock_upload_file_response
    ):
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )
        actor.init_task("Test task", max_steps=3)

        # Execute 3 steps successfully
        for _ in range(3):
            actor.step(b"image bytes")

        # 4th step should raise error
        with pytest.raises(ValueError, match="Max steps limit \\(3\\) reached"):
            actor.step(b"image bytes")

    def test_step_with_instruction(
        self, actor, sample_step, sample_usage_obj, mock_upload_file_response
    ):
        """Test that instruction parameter is accepted (though currently unused)."""
        actor.task_description = "Test task"
        actor.task_id = "existing-task"

        # Setup mocks
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        result = actor.step(b"image bytes", instruction="Click the submit button")

        # Verify chat_completion was called
        actor.client.chat_completion.assert_called_once()
        assert isinstance(result, Step)
        assert not result.stop


class TestActorContextManager:
    def test_context_manager(self, mock_sync_client):
        with Actor(api_key="test-key", base_url="https://test.example.com") as actor:
            assert actor.api_key == "test-key"
            assert actor.base_url == "https://test.example.com"

        # Verify close was called
        mock_sync_client.close.assert_called_once()

    def test_close_method(self, actor):
        actor.close()
        actor.client.close.assert_called_once()

    def test_context_manager_with_exception(self, mock_sync_client):
        try:
            with Actor(api_key="test-key", base_url="https://test.example.com"):
                raise ValueError("Test exception")
        except ValueError:
            pass

        # Verify close was still called despite exception
        mock_sync_client.close.assert_called_once()


class TestActorIntegrationScenarios:
    def test_full_workflow(
        self,
        actor,
        sample_step,
        completed_step,
        sample_usage_obj,
        mock_upload_file_response,
    ):
        """Test a complete workflow from init to completion."""
        # Initialize task
        actor.init_task("Complete workflow test")
        task_id_after_init = actor.task_id

        assert actor.task_description == "Complete workflow test"

        # Setup mocks
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response

        # First step - in progress
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )
        step1 = actor.step(b"screenshot1")
        assert not step1.stop
        assert len(step1.actions) == 1
        # task_id stays the same across steps
        assert actor.task_id == task_id_after_init

        # Second step - completed
        actor.client.chat_completion.return_value = (
            completed_step,
            "<|think_start|>done<|think_end|>\n<|action_start|>finish()<|action_end|>",
            sample_usage_obj,
        )
        step2 = actor.step(b"screenshot2")
        assert step2.stop
        assert len(step2.actions) == 1
        assert actor.task_id == task_id_after_init


class TestActorHistory:
    """Test Actor class message history functionality."""

    def test_init_task_initializes_empty_history(self, actor):
        """Test that init_task starts with empty message history."""
        actor.init_task("Test task", max_steps=5)

        assert actor.message_history == []
        assert actor.task_description == "Test task"

    def test_step_updates_message_history(
        self, actor, sample_step, sample_usage_obj, mock_upload_file_response
    ):
        """Test that step updates message_history with user and assistant messages."""
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        # First step
        actor.step(b"screenshot1")

        # Verify message_history was updated with user + assistant messages
        assert len(actor.message_history) == 2
        assert actor.message_history[0]["role"] == "user"
        assert actor.message_history[1]["role"] == "assistant"
        assert "content" in actor.message_history[0]
        assert "content" in actor.message_history[1]

    def test_step_accumulates_history_across_steps(
        self, actor, sample_step, sample_usage_obj, mock_upload_file_response
    ):
        """Test that message_history accumulates across multiple steps."""
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        # First step adds user + assistant messages
        actor.step(b"screenshot1")
        assert len(actor.message_history) == 2

        # Second step adds another user + assistant
        actor.step(b"screenshot2")
        assert len(actor.message_history) == 4

        # Should alternate user/assistant
        assert actor.message_history[0]["role"] == "user"
        assert actor.message_history[1]["role"] == "assistant"
        assert actor.message_history[2]["role"] == "user"
        assert actor.message_history[3]["role"] == "assistant"

    def test_step_only_appends_assistant_when_raw_output_exists(
        self, actor, sample_step, sample_usage_obj, mock_upload_file_response
    ):
        """Test that assistant message only added when raw_output is present."""
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_file_response

        # Return empty raw_output
        actor.client.chat_completion.return_value = (sample_step, "", sample_usage_obj)

        actor.step(b"screenshot")

        # User message is always added, but assistant message skipped when empty
        assert len(actor.message_history) == 1
        assert actor.message_history[0]["role"] == "user"


class TestActorTemperature:
    def test_task_with_default_temperature(
        self, mock_sync_client, sample_step, sample_usage_obj
    ):
        """Test that actor uses default temperature when provided."""
        mock_upload_response = UploadFileResponse(
            url="https://s3.amazonaws.com/presigned-url",
            uuid="test-uuid-123",
            expires_at=1677652888,
            file_expires_at=1677739288,
            download_url="https://cdn.example.com/test-uuid-123",
        )

        actor = Actor(
            api_key="test-key",
            base_url="https://test.example.com",
            temperature=0.5,
        )
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        actor.step(b"screenshot_data")

        # Verify temperature is passed to chat_completion
        call_args = actor.client.chat_completion.call_args
        assert call_args[1]["temperature"] == 0.5

    def test_step_temperature_overrides_task_default(
        self, mock_sync_client, sample_step, sample_usage_obj
    ):
        """Test that step temperature overrides actor default."""
        mock_upload_response = UploadFileResponse(
            url="https://s3.amazonaws.com/presigned-url",
            uuid="test-uuid-123",
            expires_at=1677652888,
            file_expires_at=1677739288,
            download_url="https://cdn.example.com/test-uuid-123",
        )

        actor = Actor(
            api_key="test-key",
            base_url="https://test.example.com",
            temperature=0.5,
        )
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        # Call step with different temperature
        actor.step(b"screenshot_data", temperature=0.9)

        # Verify step temperature (0.9) is used, not actor default (0.5)
        call_args = actor.client.chat_completion.call_args
        assert call_args[1]["temperature"] == 0.9

    def test_step_without_any_temperature(
        self, mock_sync_client, sample_step, sample_usage_obj
    ):
        """Test that when no temperature is provided, None is passed."""
        mock_upload_response = UploadFileResponse(
            url="https://s3.amazonaws.com/presigned-url",
            uuid="test-uuid-123",
            expires_at=1677652888,
            file_expires_at=1677739288,
            download_url="https://cdn.example.com/test-uuid-123",
        )

        actor = Actor(api_key="test-key", base_url="https://test.example.com")
        actor.task_description = "Test task"
        actor.client.put_s3_presigned_url.return_value = mock_upload_response
        actor.client.chat_completion.return_value = (
            sample_step,
            "<|think_start|>test<|think_end|>\n<|action_start|>click(300, 150)<|action_end|>",
            sample_usage_obj,
        )

        actor.step(b"screenshot_data")

        # Verify temperature is None (model will use its default)
        call_args = actor.client.chat_completion.call_args
        assert call_args[1]["temperature"] is None



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

from .output_parser import parse_raw_output
from .prompt_builder import build_prompt

__all__ = ["build_prompt", "parse_raw_output"]



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

"""macOS-specific keyboard and mouse input handling.

This module provides:
- macos_click(): Fix for PyAutoGUI multi-click bug on macOS
- typewrite_exact(): Type text exactly, ignoring system capslock state
"""

import time

import pyautogui

from ..exceptions import check_optional_dependency

check_optional_dependency("Quartz", "macOS multiple clicks", "desktop")
import Quartz  # noqa: E402

# macOS virtual key codes for typeable characters
KEYCODE_MAP = {
    "a": 0x00,
    "b": 0x0B,
    "c": 0x08,
    "d": 0x02,
    "e": 0x0E,
    "f": 0x03,
    "g": 0x05,
    "h": 0x04,
    "i": 0x22,
    "j": 0x26,
    "k": 0x28,
    "l": 0x25,
    "m": 0x2E,
    "n": 0x2D,
    "o": 0x1F,
    "p": 0x23,
    "q": 0x0C,
    "r": 0x0F,
    "s": 0x01,
    "t": 0x11,
    "u": 0x20,
    "v": 0x09,
    "w": 0x0D,
    "x": 0x07,
    "y": 0x10,
    "z": 0x06,
    "1": 0x12,
    "2": 0x13,
    "3": 0x14,
    "4": 0x15,
    "5": 0x17,
    "6": 0x16,
    "7": 0x1A,
    "8": 0x1C,
    "9": 0x19,
    "0": 0x1D,
    " ": 0x31,  # space
    "-": 0x1B,
    "=": 0x18,
    "[": 0x21,
    "]": 0x1E,
    "\\": 0x2A,
    ";": 0x29,
    "'": 0x27,
    "`": 0x32,
    ",": 0x2B,
    ".": 0x2F,
    "/": 0x2C,
    "\t": 0x30,  # tab
    "\n": 0x24,  # return
}

# Characters that require shift key (on US keyboard layout)
SHIFT_CHARS = set('~!@#$%^&*()_+{}|:"<>?ABCDEFGHIJKLMNOPQRSTUVWXYZ')

# Mapping of shifted characters to their base key
SHIFT_KEY_MAP = {
    "~": "`",
    "!": "1",
    "@": "2",
    "#": "3",
    "$": "4",
    "%": "5",
    "^": "6",
    "&": "7",
    "*": "8",
    "(": "9",
    ")": "0",
    "_": "-",
    "+": "=",
    "{": "[",
    "}": "]",
    "|": "\\",
    ":": ";",
    '"': "'",
    "<": ",",
    ">": ".",
    "?": "/",
}


def typewrite_exact(text: str, interval: float = 0.01) -> None:
    """Type text exactly as specified, ignoring system capslock state.

    This function uses Quartz CGEventCreateKeyboardEvent with explicit
    flag control via CGEventSetFlags() to type each character with the
    correct case, regardless of the system's capslock state.

    Args:
        text: The text to type exactly as specified
        interval: Time in seconds between each character (default: 0.01)
    """
    for char in text:
        # Determine if this character needs shift
        needs_shift = char in SHIFT_CHARS

        # Get the base key (for shifted chars, look up the unshifted version)
        if char.isupper():
            base_char = char.lower()
        elif char in SHIFT_KEY_MAP:
            base_char = SHIFT_KEY_MAP[char]
        else:
            base_char = char

        # Get keycode for the base character
        keycode = KEYCODE_MAP.get(base_char)
        if keycode is None:
            # Character not in our keycode map, skip it
            continue

        # Set flags: shift if needed, otherwise clear all flags
        flags = Quartz.kCGEventFlagMaskShift if needs_shift else 0

        # Key down
        event_down = Quartz.CGEventCreateKeyboardEvent(None, keycode, True)
        Quartz.CGEventSetFlags(event_down, flags)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, event_down)

        # Key up
        event_up = Quartz.CGEventCreateKeyboardEvent(None, keycode, False)
        Quartz.CGEventSetFlags(event_up, flags)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, event_up)

        if interval > 0:
            time.sleep(interval)


def macos_click(x: int, y: int, clicks: int = 1) -> None:
    """
    Execute a mouse click sequence on macOS with correct click state.

    This avoids the PyAutoGUI bug where multi-clicks are sent as separate
    single clicks (clickState=1), which macOS interprets as distinct events
    rather than double/triple clicks.

    Check https://github.com/asweigart/pyautogui/issues/672

    Args:
        x: X coordinate
        y: Y coordinate
        clicks: Number of clicks (1=single, 2=double, 3=triple)
    """
    # Move to position first using pyautogui to ensure consistency
    pyautogui.moveTo(x, y)

    point = Quartz.CGPoint(x=x, y=y)

    # Create and post events for each click in the sequence
    for i in range(1, clicks + 1):
        # Create Down/Up events
        mouse_down = Quartz.CGEventCreateMouseEvent(
            None, Quartz.kCGEventLeftMouseDown, point, Quartz.kCGMouseButtonLeft
        )
        mouse_up = Quartz.CGEventCreateMouseEvent(
            None, Quartz.kCGEventLeftMouseUp, point, Quartz.kCGMouseButtonLeft
        )

        # Set the click state (1 for first click, 2 for second, etc.)
        Quartz.CGEventSetIntegerValueField(
            mouse_down, Quartz.kCGMouseEventClickState, i
        )
        Quartz.CGEventSetIntegerValueField(mouse_up, Quartz.kCGMouseEventClickState, i)

        # Post events
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, mouse_down)
        Quartz.CGEventPost(Quartz.kCGHIDEventTap, mouse_up)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import logging

from .. import AsyncActor
from ..constants import (
    DEFAULT_MAX_STEPS,
    DEFAULT_STEP_DELAY,
    DEFAULT_TEMPERATURE,
    MODEL_ACTOR,
)
from ..handler.utils import configure_handler_delay, reset_handler
from ..types import (
    ActionEvent,
    AsyncActionHandler,
    AsyncImageProvider,
    AsyncObserver,
    Image,
    StepEvent,
)

logger = logging.getLogger(__name__)


def _serialize_image(image: Image | str) -> bytes | str:
    """Convert an image to bytes or keep URL as string."""
    if isinstance(image, str):
        return image
    return image.read()


class AsyncDefaultAgent:
    """Default asynchronous agent implementation using OAGI client."""

    def __init__(
        self,
        api_key: str | None = None,
        base_url: str | None = None,
        model: str = MODEL_ACTOR,
        max_steps: int = DEFAULT_MAX_STEPS,
        temperature: float | None = DEFAULT_TEMPERATURE,
        step_observer: AsyncObserver | None = None,
        step_delay: float = DEFAULT_STEP_DELAY,
    ):
        self.api_key = api_key
        self.base_url = base_url
        self.model = model
        self.max_steps = max_steps
        self.temperature = temperature
        self.step_observer = step_observer
        self.step_delay = step_delay

    async def execute(
        self,
        instruction: str,
        action_handler: AsyncActionHandler,
        image_provider: AsyncImageProvider,
    ) -> bool:
        async with AsyncActor(
            api_key=self.api_key, base_url=self.base_url, model=self.model
        ) as self.actor:
            logger.info(f"Starting async task execution: {instruction}")
            await self.actor.init_task(instruction, max_steps=self.max_steps)

            # Reset handler state at automation start
            reset_handler(action_handler)

            # Configure handler's post_batch_delay from agent's step_delay
            configure_handler_delay(action_handler, self.step_delay)

            for i in range(self.max_steps):
                step_num = i + 1
                logger.debug(f"Executing step {step_num}/{self.max_steps}")

                # Capture current state
                image = await image_provider()

                # Get next step from OAGI
                step = await self.actor.step(image, temperature=self.temperature)

                # Log reasoning
                if step.reason:
                    logger.info(f"Step {step_num}: {step.reason}")

                # Emit step event
                if self.step_observer:
                    await self.step_observer.on_event(
                        StepEvent(
                            step_num=step_num,
                            image=_serialize_image(image),
                            step=step,
                            task_id=self.actor.task_id,
                        )
                    )

                # Execute actions if any
                if step.actions:
                    logger.info(f"Actions ({len(step.actions)}):")
                    for action in step.actions:
                        count_suffix = (
                            f" x{action.count}"
                            if action.count and action.count > 1
                            else ""
                        )
                        logger.info(
                            f"  [{action.type.value}] {action.argument}{count_suffix}"
                        )

                    error = None
                    try:
                        await action_handler(step.actions)
                    except Exception as e:
                        error = str(e)
                        raise

                    # Emit action event
                    if self.step_observer:
                        await self.step_observer.on_event(
                            ActionEvent(
                                step_num=step_num,
                                actions=step.actions,
                                error=error,
                            )
                        )

                # Check if task is complete
                if step.stop:
                    logger.info(f"Task completed successfully after {step_num} steps")
                    return True

            logger.warning(
                f"Task reached max steps ({self.max_steps}) without completion"
            )
            return False



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import io
import logging
import os
import shlex
import shutil
import subprocess
import time

from screeninfo import get_monitors

from ..exceptions import check_optional_dependency
from ._ydotool import KEYCODE_MAP

check_optional_dependency("PIL", "PILImage", "desktop")
from PIL import Image  # noqa: E402

logger = logging.getLogger(__name__)


def is_wayland_display_server() -> bool:
    """Check if Wayland is the current display server."""
    return os.environ.get("WAYLAND_DISPLAY") is not None


def get_screen_size() -> tuple[int, int]:
    """Get the screen size in pixels."""
    for monitor in get_monitors():
        if monitor.is_primary:
            return monitor.width, monitor.height

    # Fallback if no monitor is marked primary
    monitors = get_monitors()
    if monitors:
        return monitors[0].width, monitors[0].height
    raise Exception("No monitor found, cannot get the screen size info")


def screenshot(region: tuple[int, int, int, int] | None = None) -> Image:
    """
    Use Flameshot to take a screenshot and return an Image object

    :return: Image object of the screenshot
    """
    # Check if flameshot is installed
    if shutil.which("flameshot") is None:
        raise RuntimeError("flameshot not found. Ensure it is installed and in PATH.")
    cmd = ["flameshot", "full", "--raw"]
    if region:
        cmd.extend(["--region", f"{region[2]}x{region[3]}+{region[0]}+{region[1]}"])
    else:
        cmd.extend(["--region", "all"])
    res = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if res.returncode != 0:
        raise RuntimeError(
            f"flameshot failed: {shlex.join(cmd)}, stdout: {res.stdout.decode(errors='ignore')}, stderr: {res.stderr.decode(errors='ignore')}"
        )
    im = Image.open(io.BytesIO(res.stdout))
    im.load()
    return im


class Ydotool:
    """
    Ydotool wrapper for Wayland display server.

    :param socket_address: The socket address for ydotool, default is empty string.
    """

    def __init__(self, socket_address: str = "") -> None:
        # Check if ydotool is installed
        if shutil.which("ydotool") is None:
            raise RuntimeError("ydotool not found. Ensure it is installed and in PATH.")
        # Set default delay between actions
        self.action_pause = 0.5
        # Last action time
        self.last_action_time = 0.0
        # Customize the socket address for ydotool
        self.socket_address = socket_address
        # Check environment issues for ydotool
        self.environ_check()

    def environ_check(self):
        """Check environment issues for ydotool"""
        # Check if ydotoold is running
        if not subprocess.run(
            ["pgrep", "ydotoold"], capture_output=True, text=True
        ).stdout.strip():
            logger.warning("Ydotool daemon (ydotoold) is not running")
        # Check the permission to access the socket address
        socket_address = (
            self.socket_address
            or os.environ.get("YDOTOOL_SOCKET", "")
            or f"/run/user/{os.getuid()}/.ydotool_socket"
        )
        if not os.access(socket_address, os.W_OK) or not os.path.exists(socket_address):
            logger.warning(f"Ydotool cannot connect to socket address:{socket_address}")
        # Check if the mouse acceleration profile is 'flat' (For GNOME)
        accel_profile = subprocess.run(
            [
                "gsettings",
                "get",
                "org.gnome.desktop.peripherals.mouse",
                "accel-profile",
            ],
            capture_output=True,
            text=True,
        ).stdout.strip()
        if accel_profile and accel_profile != "'flat'":
            logger.warning(
                f"Mouse Acceleration is not disabled, current accel-profile is {accel_profile}). Ydotool may not work as expected."
                + "Please disable mouse acceleration by running 'gsettings set org.gnome.desktop.peripherals.mouse accel-profile 'flat''",
            )

    def _get_keycode(self, key_char: str) -> int:
        """
        Get the keycode from input-event-codes mapping.
        :param key_char: Key char (e.g., "A", "ENTER", "F1", "PRINT_SCREEN", case-insensitive)
        :return: Decimal keycode
        """
        # Lookup and return keycode
        if key_char in KEYCODE_MAP:
            return KEYCODE_MAP[key_char]
        else:
            return None

    def _run_ydotool(self, args: list[str], count: int = 1) -> None:
        """
        Run ydotool command; e.g., ["click", "500", "300"] => ydotool click 500 300
        """
        if (interval := (time.time() - self.last_action_time)) < self.action_pause:
            time.sleep(interval)
        if count > 1:
            args.extend(["--repeat", str(count)])
        cmd = ["ydotool", *args]
        # Use shlex.join for clear logging
        logger.debug(f"[ydotool] {shlex.join(cmd)}")
        # Env with socket address
        env = os.environ.copy()
        if self.socket_address:
            env["YDOTOOL_SOCKET"] = self.socket_address
        # Run ydotool command
        res = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        if res.returncode != 0:
            raise RuntimeError(
                f"ydotool failed: {shlex.join(cmd)}, stdout: {res.stdout.decode(errors='ignore').strip()}, stderr: {res.stderr.decode(errors='ignore').strip()}"
            )
        self.last_action_time = time.time()

    def drag(self, x1: int, y1: int, x2: int, y2: int, count: int = 1) -> None:
        """
        Drag from (x1, y1) to (x2, y2).

        """
        for _ in range(count):
            self.mousemove(x1, y1)
            self._run_ydotool(["click", "0x40"])
            self.mousemove(x2, y2)
            self._run_ydotool(["click", "0x80"])

    def mousemove(self, x: int, y: int, count: int = 1) -> None:
        """
        Move mouse to (x, y).
        :param x: X coordinate of the mouse cursor
        :param y: Y coordinate of the mouse cursor
        :param count: Number of mouse move actions to perform
        """
        self._run_ydotool(
            ["mousemove", "--absolute", "-x", str(x), "-y", str(y)], count=count
        )

    def scroll(self, clicks: float) -> None:
        """
        Scroll mouse wheel in the given direction.
        :param clicks: Number of clicks to scroll, positive for up, negative for down
        """
        self._run_ydotool(
            [
                "mousemove",
                "-w",
                "--",
                "0",
                str(clicks),
            ],
        )

    def click(self, x: int, y: int, count: int = 1, right: bool = False) -> None:
        """
        Click at (x, y).
        """
        self.mousemove(x, y)
        if right:
            click_key = "0xC1"
        else:
            click_key = "0xC0"
        self._run_ydotool(["click", click_key], count=count)

    def type(self, text: str, count: int = 1) -> None:
        """
        Type the given text.
        """
        self._run_ydotool(["type", text], count=count)

    def hotkey(self, keys: list[str], count: int = 1) -> None:
        """
        Press and release the given keys.
        """
        hotkey_sequences = [
            self._get_keycode(key) for key in keys if self._get_keycode(key) is not None
        ]
        command_args = [f"{keycode}:1" for keycode in hotkey_sequences] + [
            f"{keycode}:0" for keycode in hotkey_sequences[::-1]
        ]
        self._run_ydotool(["key", *command_args], count=count)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------
"""OAGI action converter.

This module provides the OagiActionConverter for converting OAGI actions
to pyautogui command strings for remote execution.
"""

from typing import Any

from ..handler.utils import (
    parse_click_coords,
    parse_drag_coords,
    parse_scroll_coords,
)
from ..types import Action, ActionType
from .base import BaseActionConverter

# OAGI uses normalized 0-1000 coordinate space
OAGI_COORD_SIZE = 1000


class OagiActionConverter(BaseActionConverter[Action]):
    """Convert OAGI actions to pyautogui command strings.

    This converter handles:
    1. Coordinate scaling from 0-1000 space to sandbox dimensions (1920x1080)
    2. Action format conversion from OAGI Action format to pyautogui strings
    3. Key name normalization for hotkey combinations

    The output can be converted to runtime API steps via action_string_to_step().
    """

    @property
    def coord_width(self) -> int:
        return OAGI_COORD_SIZE

    @property
    def coord_height(self) -> int:
        return OAGI_COORD_SIZE

    def __call__(self, actions: list[Action]) -> list[str]:
        """Convert OAGI actions to list of pyautogui command strings.

        Extends base implementation to handle action count and finish detection.
        """
        converted: list[str] = []
        failed: list[tuple[str, str]] = []
        has_finish = False

        if not actions:
            return converted

        for action in actions:
            # Check for duplicate finish() during iteration
            is_finish = action.type == ActionType.FINISH
            if is_finish:
                if has_finish:
                    raise ValueError(
                        "Duplicate finish() detected. "
                        "Only one finish() is allowed per action sequence."
                    )
                has_finish = True

            try:
                converted.extend(self._convert_action(action))
            except Exception as e:
                action_repr = f"{action.type.value}({action.argument})"
                self._log_error(f"Failed to convert action: {action_repr}, error: {e}")
                failed.append((action_repr, str(e)))

        if not converted and actions and failed:
            raise RuntimeError(
                f"All action conversions failed ({len(failed)}/{len(actions)}): {failed}"
            )
        return converted

    def _convert_action(self, action: Action) -> list[str]:
        """Convert action to list of pyautogui command strings.

        Handles action.count for repeat support.
        """
        count = action.count or 1
        single_actions = self._convert_single_action(action)

        # Repeat the actions count times
        return single_actions * int(count)

    def _convert_single_action(self, action: Action) -> list[str]:
        """Convert a single OAGI action to pyautogui command string(s)."""
        action_type = action.type.value
        argument = (action.argument or "").strip("()")

        drag_duration = self.config.drag_duration
        scroll_amount = self.config.scroll_amount
        wait_duration = self.config.wait_duration
        hotkey_interval = self.config.hotkey_interval
        strict = self.config.strict_coordinate_validation

        if action_type == ActionType.CLICK.value:
            x, y = parse_click_coords(argument, self._coord_scaler, strict=strict)
            return [f"pyautogui.click(x={x}, y={y})"]

        if action_type == ActionType.LEFT_DOUBLE.value:
            x, y = parse_click_coords(argument, self._coord_scaler, strict=strict)
            return [f"pyautogui.doubleClick(x={x}, y={y})"]

        if action_type == ActionType.LEFT_TRIPLE.value:
            x, y = parse_click_coords(argument, self._coord_scaler, strict=strict)
            return [f"pyautogui.tripleClick(x={x}, y={y})"]

        if action_type == ActionType.RIGHT_SINGLE.value:
            x, y = parse_click_coords(argument, self._coord_scaler, strict=strict)
            return [f"pyautogui.rightClick(x={x}, y={y})"]

        if action_type == ActionType.DRAG.value:
            sx, sy, ex, ey = parse_drag_coords(
                argument, self._coord_scaler, strict=strict
            )
            return [
                f"pyautogui.moveTo({sx}, {sy})",
                f"pyautogui.dragTo({ex}, {ey}, duration={drag_duration})",
            ]

        if action_type == ActionType.HOTKEY.value:
            keys = self.parse_hotkey(argument, validate=True)
            valid_keys = [k for k in keys if k]
            if not valid_keys:
                raise ValueError(
                    f"Invalid hotkey format: '{argument}'. "
                    "Expected key names like 'ctrl+c', 'alt+tab'"
                )
            # Check if this is a caps lock key press
            if len(valid_keys) == 1 and valid_keys[0] == "capslock":
                if self.caps_manager.should_use_system_capslock():
                    return [f"pyautogui.hotkey('capslock', interval={hotkey_interval})"]
                else:
                    self.caps_manager.toggle()
                    return []  # No pyautogui command for session mode
            else:
                keys_str = ", ".join(repr(k) for k in valid_keys)
                return [f"pyautogui.hotkey({keys_str}, interval={hotkey_interval})"]

        if action_type == ActionType.TYPE.value:
            text = argument.strip("\"'")
            text = self.caps_manager.transform_text(text)
            return [f"pyautogui.typewrite({text!r})"]

        if action_type == ActionType.SCROLL.value:
            x, y, direction = parse_scroll_coords(
                argument, self._coord_scaler, strict=strict
            )
            amount = scroll_amount if direction == "up" else -scroll_amount
            return [f"pyautogui.moveTo({x}, {y})", f"pyautogui.scroll({amount})"]

        if action_type == ActionType.WAIT.value:
            try:
                seconds = float(argument) if argument else wait_duration
            except ValueError:
                raise ValueError(
                    f"Invalid wait duration: '{argument}'. "
                    "Expected numeric value in seconds."
                )
            return [f"WAIT({seconds})"]

        if action_type == ActionType.FINISH.value:
            self._log_info("Task completion action -> DONE")
            return ["DONE"]

        if action_type == ActionType.CALL_USER.value:
            self._log_info("User intervention requested")
            return []

        raise ValueError(
            f"Unknown action type: '{action_type}'. "
            "Supported: click, left_double, left_triple, right_single, drag, "
            "hotkey, type, scroll, wait, finish, call_user"
        )

    def serialize_actions(self, actions: list[Action]) -> list[dict[str, Any]]:
        """Serialize OAGI actions for trajectory logging."""
        return [
            {
                "type": action.type.value,
                "argument": action.argument,
                "count": action.count,
            }
            for action in (actions or [])
        ]



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

from .action_handler import ActionHandler
from .async_action_handler import AsyncActionHandler
from .async_image_provider import AsyncImageProvider
from .image import Image
from .image_provider import ImageProvider
from .models import (
    Action,
    ActionType,
    ImageConfig,
    Step,
    parse_coords,
    parse_drag_coords,
    parse_scroll,
)
from .step_observer import (
    ActionEvent,
    AsyncObserver,
    AsyncStepObserver,
    BaseEvent,
    ImageEvent,
    LogEvent,
    ObserverEvent,
    PlanEvent,
    SplitEvent,
    StepEvent,
)
from .url import URL, extract_uuid_from_url

__all__ = [
    "Action",
    "ActionEvent",
    "ActionType",
    "AsyncObserver",
    "AsyncStepObserver",
    "BaseEvent",
    "Image",
    "ImageConfig",
    "ImageEvent",
    "LogEvent",
    "ObserverEvent",
    "PlanEvent",
    "SplitEvent",
    "Step",
    "StepEvent",
    "ActionHandler",
    "AsyncActionHandler",
    "ImageProvider",
    "AsyncImageProvider",
    "URL",
    "extract_uuid_from_url",
    "parse_coords",
    "parse_drag_coords",
    "parse_scroll",
]



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------
"""
Example: OpenAI-Compatible Agent Loop

This example demonstrates how to build an agent loop using:
1. OAGI file upload for screenshots
2. OAGI action handler for execution
3. OAGI utilities for prompt building and action parsing
4. OpenAI-compatible API endpoint for LLM calls

Environment variables:
- OAGI_API_KEY: Your OAGI API key
- OAGI_BASE_URL: API base URL (default: https://api.agiopen.org)
"""

import os

from openai import OpenAI

from oagi import PyautoguiActionHandler, ScreenshotMaker, SyncClient
from oagi.utils.output_parser import parse_raw_output
from oagi.utils.prompt_builder import build_prompt

DEFAULT_BASE_URL = "https://api.agiopen.org"
DEFAULT_MODEL = "lux-thinker-1"


def agent_loop(task_description: str, max_steps: int = 100) -> bool:
    """
    Run an agent loop to complete a task.

    Args:
        task_description: The task to complete
        max_steps: Maximum number of steps to take

    Returns:
        True if task completed, False if max steps reached
    """
    # Get configuration from environment
    api_key = os.environ.get("OAGI_API_KEY")
    base_url = os.environ.get("OAGI_BASE_URL", DEFAULT_BASE_URL)

    if not api_key:
        raise ValueError("OAGI_API_KEY environment variable is required")

    # Initialize clients and handlers
    oagi_client = SyncClient(api_key=api_key, base_url=base_url)
    openai_client = OpenAI(api_key=api_key, base_url=f"{base_url}/v1")
    action_handler = PyautoguiActionHandler()
    image_provider = ScreenshotMaker()

    messages: list[dict] = []

    print(f"Starting task: {task_description}")
    print(f"Max steps: {max_steps}")
    print("-" * 50)

    try:
        for step_num in range(max_steps):
            # 1. Capture screenshot
            screenshot = image_provider()

            # 2. Upload to S3
            upload_resp = oagi_client.put_s3_presigned_url(screenshot)
            screenshot_url = upload_resp.download_url

            # 3. Build message (with prompt on first step)
            content = []
            if step_num == 0:
                content.append({"type": "text", "text": build_prompt(task_description)})
            content.append({"type": "image_url", "image_url": {"url": screenshot_url}})
            messages.append({"role": "user", "content": content})

            # 4. Call OpenAI-compatible API
            response = openai_client.chat.completions.create(
                model=DEFAULT_MODEL,
                messages=messages,
            )
            raw_output = response.choices[0].message.content or ""

            # 5. Add assistant message to history
            messages.append({"role": "assistant", "content": raw_output})

            # 6. Parse output using oagi utility
            step = parse_raw_output(raw_output)
            print(f"Step {step_num}: {step.reason}")
            print(f"  Actions: {step.actions}")

            # 7. Check for completion
            if step.stop:
                print("-" * 50)
                print("Task completed!")
                return True

            # 8. Execute actions
            action_handler(step.actions)

        print("-" * 50)
        print("Max steps reached without completion")
        return False

    finally:
        oagi_client.close()
        openai_client.close()


if __name__ == "__main__":
    # Example task
    task = "Find some shoes on amazon"

    success = agent_loop(task, max_steps=100)
    print(f"\nFinal result: {'Success' if success else 'Failed'}")



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

"""Windows-specific keyboard input handling.

This module provides typewrite_exact() which types text exactly as specified,
ignoring the system's capslock state by using SendInput with KEYEVENTF_UNICODE.
"""

import ctypes
import time
from ctypes import wintypes

INPUT_KEYBOARD = 1
KEYEVENTF_UNICODE = 0x0004
KEYEVENTF_KEYUP = 0x0002


class KEYBDINPUT(ctypes.Structure):
    _fields_ = [
        ("wVk", wintypes.WORD),
        ("wScan", wintypes.WORD),
        ("dwFlags", wintypes.DWORD),
        ("time", wintypes.DWORD),
        ("dwExtraInfo", ctypes.POINTER(ctypes.c_ulong)),
    ]


class MOUSEINPUT(ctypes.Structure):
    _fields_ = [
        ("dx", ctypes.c_long),
        ("dy", ctypes.c_long),
        ("mouseData", wintypes.DWORD),
        ("dwFlags", wintypes.DWORD),
        ("time", wintypes.DWORD),
        ("dwExtraInfo", ctypes.POINTER(ctypes.c_ulong)),
    ]


class HARDWAREINPUT(ctypes.Structure):
    _fields_ = [
        ("uMsg", wintypes.DWORD),
        ("wParamL", wintypes.WORD),
        ("wParamH", wintypes.WORD),
    ]


class INPUT(ctypes.Structure):
    class _I(ctypes.Union):
        _fields_ = [
            ("ki", KEYBDINPUT),
            ("mi", MOUSEINPUT),
            ("hi", HARDWAREINPUT),
        ]

    _anonymous_ = ("i",)
    _fields_ = [
        ("type", wintypes.DWORD),
        ("i", _I),
    ]


# Configure SendInput with proper argtypes for 64-bit compatibility
SendInput = ctypes.windll.user32.SendInput
SendInput.argtypes = [wintypes.UINT, ctypes.POINTER(INPUT), ctypes.c_int]
SendInput.restype = wintypes.UINT


def typewrite_exact(text: str, interval: float = 0.01) -> None:
    """Type text exactly using Unicode input - ignores capslock, keyboard layout, etc.

    This function uses SendInput with KEYEVENTF_UNICODE to send characters
    directly by their Unicode codepoint, completely bypassing keyboard state
    (capslock, layout, etc.).

    Args:
        text: The text to type exactly as specified
        interval: Time in seconds between each character (default: 0.01)
    """
    for char in text:
        inputs = (INPUT * 2)()

        # Key down
        inputs[0].type = INPUT_KEYBOARD
        inputs[0].ki.wScan = ord(char)
        inputs[0].ki.dwFlags = KEYEVENTF_UNICODE

        # Key up
        inputs[1].type = INPUT_KEYBOARD
        inputs[1].ki.wScan = ord(char)
        inputs[1].ki.dwFlags = KEYEVENTF_UNICODE | KEYEVENTF_KEYUP

        SendInput(2, inputs, ctypes.sizeof(INPUT))

        if interval > 0:
            time.sleep(interval)



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------

import pytest

from oagi.types import parse_coords, parse_drag_coords, parse_scroll


class TestParseCoords:
    @pytest.mark.parametrize(
        "input_str,expected",
        [
            ("500, 300", (500, 300)),
            ("500,300,", (500, 300)),
            ("0, 0", (0, 0)),
            ("1000, 1000", (1000, 1000)),
            ("123,456", (123, 456)),
            ("  500 ,  300  ", None),  # spaces before digits not supported
        ],
    )
    def test_valid_coords(self, input_str, expected):
        assert parse_coords(input_str) == expected

    @pytest.mark.parametrize(
        "input_str",
        [
            "",
            "invalid",
            "500",
            "500,",
            ",300",
            "abc, def",
            "500.5, 300.5",
        ],
    )
    def test_invalid_coords(self, input_str):
        assert parse_coords(input_str) is None


class TestParseDragCoords:
    @pytest.mark.parametrize(
        "input_str,expected",
        [
            ("100, 100, 500, 500", (100, 100, 500, 500)),
            ("100,100,500,500,", (100, 100, 500, 500)),
            ("0, 0, 1000, 1000", (0, 0, 1000, 1000)),
            ("50,50,200,300", (50, 50, 200, 300)),
        ],
    )
    def test_valid_drag_coords(self, input_str, expected):
        assert parse_drag_coords(input_str) == expected

    @pytest.mark.parametrize(
        "input_str",
        [
            "",
            "invalid",
            "100, 100, 500",
            "100, 100",
            "100",
            "abc, def, ghi, jkl",
        ],
    )
    def test_invalid_drag_coords(self, input_str):
        assert parse_drag_coords(input_str) is None


class TestParseScroll:
    @pytest.mark.parametrize(
        "input_str,expected",
        [
            ("500, 300, up", (500, 300, "up")),
            ("500,300,down,", (500, 300, "down")),
            ("500, 300, UP", (500, 300, "up")),
            ("500, 300, Down", (500, 300, "down")),
            ("0, 0, up", (0, 0, "up")),
        ],
    )
    def test_valid_scroll(self, input_str, expected):
        assert parse_scroll(input_str) == expected

    @pytest.mark.parametrize(
        "input_str",
        [
            "",
            "invalid",
            "500, 300",
            "500",
            "500, 300,",
            "500, 300, left",  # invalid direction
            "500, 300, right",  # invalid direction
            "500, 300, invalid_direction",  # invalid direction
        ],
    )
    def test_invalid_scroll(self, input_str):
        assert parse_scroll(input_str) is None



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------
"""Base class for action converters.

This module provides the abstract base class for converting model-specific
actions to pyautogui command strings for remote execution.
"""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from ..handler.capslock_manager import CapsLockManager
from ..handler.utils import (
    CoordinateScaler,
    normalize_key,
    parse_hotkey,
    validate_keys,
)

T = TypeVar("T")


@dataclass
class ConverterConfig:
    """Configuration for action converters.

    Matches the configuration options in PyautoguiConfig for consistency.
    """

    sandbox_width: int = 1920
    sandbox_height: int = 1080
    drag_duration: float = 0.5
    scroll_amount: int = 2
    wait_duration: float = 1.0
    hotkey_interval: float = 0.1
    capslock_mode: str = "session"
    strict_coordinate_validation: bool = False
    """If True, raise ValueError when coordinates are outside valid range.
    If False (default), clamp coordinates to valid range (original behavior)."""


class BaseActionConverter(ABC, Generic[T]):
    """Abstract base class for action converters.

    Subclasses must implement:
    - coord_width/coord_height properties for input coordinate space
    - _convert_single_action() for model-specific conversion logic
    - serialize_actions() for trajectory logging

    Provides common functionality:
    - Coordinate scaling via CoordinateScaler
    - Key normalization via shared utils
    - __call__ interface returning list of action strings
    - action_string_to_step() for runtime API format
    """

    def __init__(
        self,
        *,
        config: ConverterConfig | None = None,
        logger: Any | None = None,
    ):
        """Initialize the converter.

        Args:
            config: Converter configuration. Uses defaults if not provided.
            logger: Optional logger instance for debug/error logging.
        """
        self.config = config or ConverterConfig()
        self.logger = logger

        # Initialize coordinate scaler
        self._coord_scaler = CoordinateScaler(
            source_width=self.coord_width,
            source_height=self.coord_height,
            target_width=self.config.sandbox_width,
            target_height=self.config.sandbox_height,
        )

        # Initialize caps lock manager
        self.caps_manager = CapsLockManager(mode=self.config.capslock_mode)

        # Track last cursor position (for actions without explicit coordinates)
        self._last_x: int | None = None
        self._last_y: int | None = None

    @property
    @abstractmethod
    def coord_width(self) -> int:
        """Input coordinate space width (e.g., 1024 for XGA, 1000 for OAGI)."""
        ...

    @property
    @abstractmethod
    def coord_height(self) -> int:
        """Input coordinate space height (e.g., 768 for XGA, 1000 for OAGI)."""
        ...

    @property
    def scale_x(self) -> float:
        """X scaling factor from input to sandbox coordinates."""
        return self._coord_scaler.scale_x

    @property
    def scale_y(self) -> float:
        """Y scaling factor from input to sandbox coordinates."""
        return self._coord_scaler.scale_y

    def scale_coordinate(self, x: int | float, y: int | float) -> tuple[int, int]:
        """Scale coordinates from model space to sandbox space.

        Args:
            x: X coordinate in model space
            y: Y coordinate in model space

        Returns:
            Tuple of (scaled_x, scaled_y) in sandbox space
        """
        return self._coord_scaler.scale(x, y)

    def normalize_key(self, key: str) -> str:
        """Normalize a key name to pyautogui format.

        Args:
            key: Key name to normalize

        Returns:
            Normalized key name
        """
        return normalize_key(key)

    def parse_hotkey(self, hotkey_str: str, *, validate: bool = True) -> list[str]:
        """Parse a hotkey string into a list of normalized key names.

        Args:
            hotkey_str: Hotkey string (e.g., "ctrl+c")
            validate: If True, validate keys against PYAUTOGUI_VALID_KEYS

        Returns:
            List of normalized key names
        """
        return parse_hotkey(hotkey_str, validate=validate)

    def validate_keys(self, keys: list[str]) -> None:
        """Validate that all keys are recognized by pyautogui.

        Args:
            keys: List of key names to validate

        Raises:
            ValueError: If any key is invalid
        """
        validate_keys(keys)

    def _get_last_or_center(self) -> tuple[int, int]:
        """Get last cursor position or screen center as fallback.

        Returns:
            Tuple of (x, y) coordinates
        """
        if self._last_x is not None and self._last_y is not None:
            return self._last_x, self._last_y
        return self.config.sandbox_width // 2, self.config.sandbox_height // 2

    def _log_error(self, message: str) -> None:
        """Log an error message if logger is available."""
        if self.logger:
            self.logger.error(message)

    def _log_info(self, message: str) -> None:
        """Log an info message if logger is available."""
        if self.logger:
            self.logger.info(message)

    def _log_debug(self, message: str) -> None:
        """Log a debug message if logger is available."""
        if self.logger:
            self.logger.debug(message)

    def __call__(self, actions: list[T]) -> list[str]:
        """Convert actions to list of pyautogui command strings.

        Args:
            actions: List of model-specific action objects

        Returns:
            List of pyautogui command strings

        Raises:
            RuntimeError: If all action conversions failed
        """
        converted: list[str] = []
        failed: list[tuple[str, str]] = []
        skipped: list[str] = []

        if not actions:
            return converted

        for action in actions:
            try:
                action_strings = self._convert_single_action(action)

                if not action_strings:
                    # No-op action (e.g., screenshot, cursor_position)
                    action_type = getattr(action, "action_type", repr(action))
                    skipped.append(str(action_type))
                    continue

                converted.extend(action_strings)

            except Exception as e:
                action_repr = repr(action)
                self._log_error(f"Failed to convert action: {action_repr}, error: {e}")
                failed.append((action_repr, str(e)))

        if skipped:
            self._log_debug(f"Skipped no-op actions: {skipped}")

        if not converted and actions and failed:
            raise RuntimeError(
                f"All action conversions failed ({len(failed)}/{len(actions)}): {failed}"
            )

        return converted

    @abstractmethod
    def _convert_single_action(self, action: T) -> list[str]:
        """Convert a single action to pyautogui command string(s).

        Args:
            action: Model-specific action object

        Returns:
            List of pyautogui command strings (may be empty for no-op actions)

        Raises:
            ValueError: If action format is invalid
        """
        ...

    @abstractmethod
    def serialize_actions(self, actions: list[T]) -> list[dict[str, Any]]:
        """Serialize actions for trajectory logging.

        Args:
            actions: List of model-specific action objects

        Returns:
            List of serialized action dictionaries
        """
        ...

    def action_string_to_step(self, action: str) -> dict[str, Any]:
        """Convert an action string into a step for runtime/do API.

        Args:
            action: Action string (e.g., "pyautogui.click(x=100, y=200)")

        Returns:
            Step dict for runtime API
        """
        action_str = str(action).strip()

        # Special markers
        upper = action_str.upper()
        if upper in ["DONE", "FAIL"]:
            return {"type": "sleep", "parameters": {"seconds": 0}}

        # WAIT(seconds)
        wait_match = re.match(
            r"^WAIT\((?P<sec>[0-9]*\.?[0-9]+)\)$", action_str, re.IGNORECASE
        )
        if wait_match:
            seconds = float(wait_match.group("sec"))
            return {"type": "sleep", "parameters": {"seconds": seconds}}

        # pyautogui code path
        if "pyautogui" in action_str.lower():
            return {
                "type": "pyautogui",
                "parameters": {"code": action_str},
            }

        # Default: shell command
        return {"type": "execute", "parameters": {"command": action_str, "shell": True}}



# -----------------------------------------------------------------------------
#  Copyright (c) OpenAGI Foundation
#  All rights reserved.
#
#  This file is part of the official API project.
#  Licensed under the MIT License.
# -----------------------------------------------------------------------------
"""Shared utilities for action handling and conversion.

This module provides common functionality used by both PyautoguiActionHandler
(for local execution) and action converters (for remote execution).
"""

import sys

# =============================================================================
# Key Normalization Mapping
# =============================================================================

# Minimal key mapping - only normalizes common variations to pyautogui names
# Matches original PyautoguiActionHandler.hotkey_variations_mapping behavior exactly:
#   "capslock": ["caps_lock", "caps", "capslock"] -> capslock
#   "pgup": ["page_up", "pageup"] -> pgup
#   "pgdn": ["page_down", "pagedown"] -> pgdn
KEY_MAP: dict[str, str] = {
    # Caps lock variations -> capslock
    "caps_lock": "capslock",
    "caps": "capslock",
    # Page up variations -> pgup (short form, matching original)
    "page_up": "pgup",
    "pageup": "pgup",
    # Page down variations -> pgdn (short form, matching original)
    "page_down": "pgdn",
    "pagedown": "pgdn",
}

# Valid pyautogui key names
PYAUTOGUI_VALID_KEYS: frozenset[str] = frozenset(
    {
        # Alphabet keys
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
        # Number keys
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        # Function keys
        "f1",
        "f2",
        "f3",
        "f4",
        "f5",
        "f6",
        "f7",
        "f8",
        "f9",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
        "f19",
        "f20",
        "f21",
        "f22",
        "f23",
        "f24",
        # Navigation keys
        "up",
        "down",
        "left",
        "right",
        "home",
        "end",
        "pageup",
        "pagedown",
        "pgup",
        "pgdn",
        # Editing keys
        "backspace",
        "delete",
        "del",
        "insert",
        "enter",
        "return",
        "tab",
        "space",
        # Modifier keys (with left/right variants)
        "shift",
        "shiftleft",
        "shiftright",
        "ctrl",
        "ctrlleft",
        "ctrlright",
        "alt",
        "altleft",
        "altright",
        "option",
        "optionleft",
        "optionright",
        "command",
        "win",
        "winleft",
        "winright",
        "fn",
        # Lock keys
        "capslock",
        "numlock",
        "scrolllock",
        # Special keys
        "esc",
        "escape",
        "pause",
        "printscreen",
        "prtsc",
        "prtscr",
        "prntscrn",
        "print",
        "apps",
        "clear",
        "sleep",
        # Symbols
        "!",
        "@",
        "#",
        "$",
        "%",
        "^",
        "&",
        "*",
        "(",
        ")",
        "-",
        "_",
        "=",
        "+",
        "[",
        "]",
        "{",
        "}",
        "\\",
        "|",
        ";",
        ":",
        "'",
        '"',
        ",",
        ".",
        "<",
        ">",
        "/",
        "?",
        "`",
        "~",
        # Numpad keys
        "num0",
        "num1",
        "num2",
        "num3",
        "num4",
        "num5",
        "num6",
        "num7",
        "num8",
        "num9",
        "divide",
        "multiply",
        "subtract",
        "add",
        "decimal",
        # Media keys
        "volumeup",
        "volumedown",
        "volumemute",
        "playpause",
        "stop",
        "nexttrack",
        "prevtrack",
        # Browser keys
        "browserback",
        "browserforward",
        "browserrefresh",
        "browserstop",
        "browsersearch",
        "browserfavorites",
        "browserhome",
        # Application launch keys
        "launchapp1",
        "launchapp2",
        "launchmail",
        "launchmediaselect",
    }
)


# =============================================================================
# Coordinate Scaling
# =============================================================================


class CoordinateScaler:
    """Handles coordinate scaling between different coordinate systems.

    This class provides reusable coordinate transformation logic used by both
    PyautoguiActionHandler (local execution) and action converters (remote execution).

    Args:
        source_width: Width of the source coordinate space (e.g., 1000 for OAGI)
        source_height: Height of the source coordinate space
        target_width: Width of the target coordinate space (e.g., screen width)
        target_height: Height of the target coordinate space
        origin_x: X offset of the target coordinate origin (for multi-monitor)
        origin_y: Y offset of the target coordinate origin (for multi-monitor)
    """

    def __init__(
        self,
        source_width: int,
        source_height: int,
        target_width: int,
        target_height: int,
        origin_x: int = 0,
        origin_y: int = 0,
    ):
        self.source_width = source_width
        self.source_height = source_height
        self.target_width = target_width
        self.target_height = target_height
        self.origin_x = origin_x
        self.origin_y = origin_y
        self.scale_x = target_width / source_width
        self.scale_y = target_height / source_height

    def scale(
        self,
        x: int | float,
        y: int | float,
        *,
        clamp: bool = True,
        prevent_failsafe: bool = False,
        strict: bool = False,
    ) -> tuple[int, int]:
        """Scale coordinates from source to target space.

        Args:
            x: X coordinate in source space
            y: Y coordinate in source space
            clamp: If True, clamp to valid target range
            prevent_failsafe: If True, offset corner coordinates by 1 pixel
                             (prevents PyAutoGUI fail-safe trigger)
            strict: If True, raise ValueError when coordinates are outside
                   valid source range [0, source_width] x [0, source_height]

        Returns:
            Tuple of (target_x, target_y) in target coordinate space

        Raises:
            ValueError: If strict=True and coordinates are outside valid range
        """
        # Strict validation: check if coordinates are in valid source range
        if strict:
            if x < 0 or x > self.source_width:
                raise ValueError(
                    f"x coordinate {x} out of valid range [0, {self.source_width}]. "
                    f"Coordinates must be normalized between 0 and {self.source_width}."
                )
            if y < 0 or y > self.source_height:
                raise ValueError(
                    f"y coordinate {y} out of valid range [0, {self.source_height}]. "
                    f"Coordinates must be normalized between 0 and {self.source_height}."
                )

        scaled_x = round(x * self.scale_x)
        scaled_y = round(y * self.scale_y)

        if clamp:
            # Clamp to valid range
            scaled_x = max(0, min(scaled_x, self.target_width - 1))
            scaled_y = max(0, min(scaled_y, self.target_height - 1))

        if prevent_failsafe:
            # Prevent PyAutoGUI fail-safe by adjusting corner coordinates
            if scaled_x < 1:
                scaled_x = 1
            elif scaled_x > self.target_width - 2:
                scaled_x = self.target_width - 2
            if scaled_y < 1:
                scaled_y = 1
            elif scaled_y > self.target_height - 2:
                scaled_y = self.target_height - 2

        # Add origin offset (for multi-monitor support)
        return scaled_x + self.origin_x, scaled_y + self.origin_y

    def set_origin(self, origin_x: int, origin_y: int) -> None:
        """Update the origin offset."""
        self.origin_x = origin_x
        self.origin_y = origin_y

    def set_target_size(self, width: int, height: int) -> None:
        """Update the target size and recalculate scale factors."""
        self.target_width = width
        self.target_height = height
        self.scale_x = width / self.source_width
        self.scale_y = height / self.source_height


# =============================================================================
# Key Normalization Functions
# =============================================================================


def normalize_key(key: str, *, macos_ctrl_to_cmd: bool = False) -> str:
    """Normalize a key name to pyautogui format.

    Args:
        key: Key name to normalize (e.g., "ctrl", "Control", "page_down")
        macos_ctrl_to_cmd: If True and on macOS, remap 'ctrl' to 'command'

    Returns:
        Normalized key name (e.g., "ctrl", "pagedown")
    """
    key = key.strip().lower()
    normalized = KEY_MAP.get(key, key)

    # Remap ctrl to command on macOS if enabled
    if macos_ctrl_to_cmd and sys.platform == "darwin" and normalized == "ctrl":
        return "command"

    return normalized


def parse_hotkey(
    hotkey_str: str,
    *,
    macos_ctrl_to_cmd: bool = False,
    validate: bool = True,
) -> list[str]:
    """Parse a hotkey string into a list of normalized key names.

    Args:
        hotkey_str: Hotkey string (e.g., "ctrl+c", "alt, tab", "Shift+Enter")
        macos_ctrl_to_cmd: If True and on macOS, remap 'ctrl' to 'command'
        validate: If True, validate keys against PYAUTOGUI_VALID_KEYS

    Returns:
        List of normalized key names (e.g., ["ctrl", "c"])

    Raises:
        ValueError: If validate=True and any key is invalid
    """
    # Remove parentheses if present
    hotkey_str = hotkey_str.strip("()")

    # Split by '+' or ',' to get individual keys
    if "+" in hotkey_str:
        keys = [
            normalize_key(k, macos_ctrl_to_cmd=macos_ctrl_to_cmd)
            for k in hotkey_str.split("+")
        ]
    else:
        keys = [
            normalize_key(k, macos_ctrl_to_cmd=macos_ctrl_to_cmd)
            for k in hotkey_str.split(",")
        ]

    # Filter empty strings
    keys = [k for k in keys if k]

    if validate:
        validate_keys(keys)

    return keys


def validate_keys(keys: list[str]) -> None:
    """Validate that all keys are recognized by pyautogui.

    Args:
        keys: List of normalized key names

    Raises:
        ValueError: If any key is invalid, with helpful suggestions
    """
    invalid_keys = [k for k in keys if k and k not in PYAUTOGUI_VALID_KEYS]

    if invalid_keys:
        suggestions = []
        for invalid_key in invalid_keys:
            if invalid_key in ("ret",):
                suggestions.append(f"'{invalid_key}' -> use 'enter' or 'return'")
            elif invalid_key.startswith("num") and len(invalid_key) > 3:
                suggestions.append(
                    f"'{invalid_key}' -> numpad keys use format 'num0'-'num9'"
                )
            else:
                suggestions.append(f"'{invalid_key}' is not a valid key name")

        error_msg = "Invalid key name(s) in hotkey: " + ", ".join(suggestions)
        valid_sample = ", ".join(sorted(list(PYAUTOGUI_VALID_KEYS)[:30]))
        error_msg += f"\n\nValid keys include: {valid_sample}... (and more)"
        raise ValueError(error_msg)


# =============================================================================
# Coordinate Parsing Functions
# =============================================================================


def parse_click_coords(
    argument: str,
    scaler: CoordinateScaler,
    *,
    prevent_failsafe: bool = False,
    strict: bool = False,
) -> tuple[int, int]:
    """Parse click coordinates from argument string.

    Args:
        argument: Coordinate string in format "x, y"
        scaler: CoordinateScaler instance for coordinate transformation
        prevent_failsafe: If True, offset corner coordinates
        strict: If True, raise ValueError for out-of-range coordinates

    Returns:
        Tuple of (x, y) in target coordinate space

    Raises:
        ValueError: If coordinate format is invalid or (strict=True) out of range
    """
    # Check for common format errors
    if " and " in argument.lower() or " then " in argument.lower():
        raise ValueError(
            f"Invalid click format: '{argument}'. "
            "Cannot combine multiple actions with 'and' or 'then'."
        )

    parts = argument.split(",") if argument else []
    if len(parts) < 2:
        raise ValueError(
            f"Invalid click coordinate format: '{argument}'. "
            "Expected 'x, y' (comma-separated numeric values)"
        )

    try:
        x = float(parts[0].strip())
        y = float(parts[1].strip())
        return scaler.scale(x, y, prevent_failsafe=prevent_failsafe, strict=strict)
    except (ValueError, IndexError) as e:
        raise ValueError(
            f"Failed to parse click coords '{argument}': {e}. "
            "Coordinates must be comma-separated numeric values."
        ) from e


def parse_drag_coords(
    argument: str,
    scaler: CoordinateScaler,
    *,
    prevent_failsafe: bool = False,
    strict: bool = False,
) -> tuple[int, int, int, int]:
    """Parse drag coordinates from argument string.

    Args:
        argument: Coordinate string in format "x1, y1, x2, y2"
        scaler: CoordinateScaler instance for coordinate transformation
        prevent_failsafe: If True, offset corner coordinates
        strict: If True, raise ValueError for out-of-range coordinates

    Returns:
        Tuple of (x1, y1, x2, y2) in target coordinate space

    Raises:
        ValueError: If coordinate format is invalid or (strict=True) out of range
    """
    # Check for common format errors
    if " and " in argument.lower() or " then " in argument.lower():
        raise ValueError(
            f"Invalid drag format: '{argument}'. "
            "Cannot combine multiple actions with 'and' or 'then'."
        )

    parts = argument.split(",") if argument else []
    if len(parts) != 4:
        raise ValueError(
            f"Invalid drag coordinate format: '{argument}'. "
            "Expected 'x1, y1, x2, y2' (4 comma-separated numeric values)"
        )

    try:
        sx = float(parts[0].strip())
        sy = float(parts[1].strip())
        ex = float(parts[2].strip())
        ey = float(parts[3].strip())
        x1, y1 = scaler.scale(sx, sy, prevent_failsafe=prevent_failsafe, strict=strict)
        x2, y2 = scaler.scale(ex, ey, prevent_failsafe=prevent_failsafe, strict=strict)
        return x1, y1, x2, y2
    except (ValueError, IndexError) as e:
        raise ValueError(
            f"Failed to parse drag coords '{argument}': {e}. "
            "Coordinates must be comma-separated numeric values."
        ) from e


def parse_scroll_coords(
    argument: str,
    scaler: CoordinateScaler,
    *,
    prevent_failsafe: bool = False,
    strict: bool = False,
) -> tuple[int, int, str]:
    """Parse scroll coordinates and direction from argument string.

    Args:
        argument: Scroll string in format "x, y, direction"
        scaler: CoordinateScaler instance for coordinate transformation
        prevent_failsafe: If True, offset corner coordinates
        strict: If True, raise ValueError for out-of-range coordinates

    Returns:
        Tuple of (x, y, direction) where direction is 'up' or 'down'

    Raises:
        ValueError: If format is invalid or (strict=True) coordinates out of range
    """
    parts = [p.strip() for p in argument.split(",")]
    if len(parts) != 3:
        raise ValueError(
            f"Invalid scroll format: '{argument}'. "
            "Expected 'x, y, direction' (e.g., '500, 300, up')"
        )

    try:
        x = float(parts[0])
        y = float(parts[1])
        direction = parts[2].lower()

        if direction not in ("up", "down"):
            raise ValueError(
                f"Invalid scroll direction: '{direction}'. Use 'up' or 'down'."
            )

        scaled_x, scaled_y = scaler.scale(
            x, y, prevent_failsafe=prevent_failsafe, strict=strict
        )
        return scaled_x, scaled_y, direction
    except (ValueError, IndexError) as e:
        if "scroll direction" in str(e):
            raise
        raise ValueError(
            f"Failed to parse scroll coords '{argument}': {e}. "
            "Format: 'x, y, direction'"
        ) from e


# =============================================================================
# Handler Utility Functions
# =============================================================================


def reset_handler(handler) -> None:
    """Reset handler state if supported.

    Uses duck-typing to check if the handler has a reset() method.
    This allows handlers to reset their internal state (e.g., capslock state)
    at the start of a new automation task.

    Args:
        handler: The action handler to reset
    """
    if hasattr(handler, "reset"):
        handler.reset()


def configure_handler_delay(handler, step_delay: float) -> None:
    """Configure handler's post_batch_delay from agent's step_delay.

    Uses duck-typing to check if the handler has a config with post_batch_delay.
    This allows agents to control the delay after action execution.

    Args:
        handler: The action handler to configure
        step_delay: The delay in seconds to set
    """
    if hasattr(handler, "config") and hasattr(handler.config, "post_batch_delay"):
        handler.config.post_batch_delay = step_delay
